"""Python Runsolver class."""

import argparse
import os
import pty
import select
import shlex
import subprocess
import sys
import time
from typing import Optional
from pathlib import Path
import psutil

from sparkle.tools.general import get_time_pid_random_string
from sparkle.tools.runsolver.runsolver import RunSolver
from sparkle.__about__ import __version__ as sparkle_version


class PyRunSolver(RunSolver):
    """Class representation of Python based RunSolver."""

    @staticmethod
    def wrap_command(
        command: list[str],
        cutoff_time: int,
        log_directory: Path,
        log_name_base: str = None,
        raw_results_file: bool = True,
    ) -> list[str]:
        """Wrap a command with the RunSolver call and arguments.

        Args:
            command: The command to wrap.
            cutoff_time: The cutoff CPU time for the solver.
            log_directory: The directory where to write the solver output.
            log_name_base: A user defined name to easily identify the logs.
                Defaults to "runsolver".
            raw_results_file: Whether to use the raw results file.

        Returns:
            List of commands and arguments to execute the solver.
        """
        log_name_base = "runsolver" if log_name_base is None else log_name_base
        unique_stamp = get_time_pid_random_string()
        raw_result_path = log_directory / Path(f"{log_name_base}_{unique_stamp}.rawres")
        watcher_data_path = raw_result_path.with_suffix(".log")
        var_values_path = raw_result_path.with_suffix(".val")
        runner_exec = Path(__file__)

        return (
            [
                sys.executable,
                str(runner_exec.absolute()),
                "--timestamp",
                "-C",
                str(cutoff_time),
                "-w",
                str(watcher_data_path),
                "-v",
                str(var_values_path),
            ]
            + (["-o", str(raw_result_path)] if raw_results_file else [])
            + command
        )


def run_with_monitoring(
    command: list[str],
    watcher_file: Path,
    value_file: Path,
    output_file: Path,
    timestamp: bool = True,
    cpu_limit: Optional[int] = None,
    wall_clock_limit: Optional[int] = None,
    vm_limit: Optional[int] = None,
) -> None:
    """Runs a command with CPU, wall-clock, and memory monitoring.

    Args:
        command: The command to execute as a list of strings.
        watcher_file: File to log the command line.
        value_file: File to write final resource usage metrics.
        output_file: Optional file to redirect command's output.
        timestamp: Whether to add timestamps to each raw output line as CPU TIME / WC Time.
        cpu_limit: CPU time limit in seconds.
        wall_clock_limit: Wall-clock time limit in seconds.
        vm_limit: Virtual memory limit in KiB.
    """
    start_time = time.time()
    cpu_time = 0.0
    user_time = 0.0
    system_time = 0.0
    max_memory_kib = 0

    with watcher_file.open("w") as f:
        f.write(
            f"PyRunSolver from Sparkle v{sparkle_version}, a Python mirror of RunSolver. Copyright (C) 2025, ADA Research Group\n"
        )
        f.write(f"command line: {shlex.join(sys.argv)}")

    def process_raw_output(fd: int, target_file: Path = None) -> int | None:
        """Reads and writes 'raw' command output to a file."""
        if fd is None:  # Closed file stream, nothing to do
            return None
        while select.select([fd], [], [], 0)[
            0
        ]:  # Check if FD is available for read without blocking/waiting
            try:
                data = os.read(
                    fd,
                    8192,  # Preferably this would be larger, but the stream never gives more than 4095 chars per read. Don't know why.
                ).decode()
                if data:
                    if timestamp:
                        ends_with_newline = data.endswith("\n")
                        stamp = f"{user_time:.2f}/{time.time() - start_time:.2f}"
                        data = "\n".join(
                            [f"{stamp}\t{line}" for line in data.splitlines()]
                        )  # Add stamp at the beginning of each line
                        data += (
                            "\n" if ends_with_newline else ""
                        )  # Splitlines removes last \n
                    if target_file:
                        with target_file.open(
                            "a"
                        ) as f:  # Reopen and close per line to stream output
                            f.write(data)
                    else:  # No output log, print to 'terminal' (Or slurm log etc.)
                        print(data)
                else:
                    os.close(fd)  # No data on stream
                    return None  # Remove FD
            except OSError:  # Streamno longer readable
                return None  # Remove FD
        return fd

    if output_file:  # Create raw output log
        output_file.open("w+").close()

    try:
        master_fd, slave_fd = (
            pty.openpty()
        )  # Open new pseudo-terminal pair for Master (us) and Slave (subprocess)
        process = subprocess.Popen(
            command, stdout=slave_fd, stderr=slave_fd, close_fds=True
        )
        os.close(slave_fd)  # Is this necessary? Are the fds not already closed above?
        ps_process = psutil.Process(process.pid)

        # Main monitoring loop
        while process.poll() is None:
            # Check if process has exceed wall clock limit
            if wall_clock_limit and (time.time() - start_time) > wall_clock_limit:
                process.kill()
                break

            # Read process output and write to out_log
            master_fd = process_raw_output(master_fd, output_file)

            try:  # Try to update statistics
                children = ps_process.children(recursive=True)

                # Sum CPU times and memory for the entire process tree
                current_user_time = (
                    ps_process.cpu_times().user + ps_process.cpu_times().children_user
                )
                current_system_time = (
                    ps_process.cpu_times().system
                    + ps_process.cpu_times().children_system
                )
                current_vms = ps_process.memory_info().vms

                for child in children:
                    try:
                        current_user_time += child.cpu_times().user
                        current_system_time += child.cpu_times().system
                        current_vms += child.memory_info().vms
                    except psutil.NoSuchProcess:
                        continue

                user_time = current_user_time
                system_time = current_system_time
                cpu_time = user_time + system_time
                max_memory_kib = max(max_memory_kib, current_vms / 1024)
                if cpu_limit and cpu_time > cpu_limit:
                    process.kill()
                    break
                if vm_limit and max_memory_kib > vm_limit:
                    process.kill()
                    break

            except psutil.NoSuchProcess:
                break
        process_raw_output(master_fd, output_file)  # Final read from stream
    except (
        KeyboardInterrupt
    ):  # Ensure that we can catch CTRL-C and still wrap up properly
        if process and process.poll() is None:
            process.kill()
        raise
    finally:
        if process:
            process.wait()
        if master_fd:
            os.close(master_fd)  # Close child's output stream

    wall_time = time.time() - start_time
    timeout = (cpu_limit is not None and cpu_time > cpu_limit) or (
        wall_clock_limit is not None and wall_time > wall_clock_limit
    )
    memout = vm_limit is not None and max_memory_kib > vm_limit

    stats = {
        "WCTIME": (f"{wall_time}", "wall clock time in seconds"),
        "CPUTIME": (f"{cpu_time}", "CPU time in seconds (USERTIME+SYSTEMTIME)"),
        "USERTIME": (f"{user_time}", "CPU time spent in user mode in seconds"),
        "SYSTEMTIME": (f"{system_time}", "CPU time spent in system mode in seconds"),
        "CPUUSAGE": (
            f"{((cpu_time / wall_time) * 100 if wall_time > 0 else 0):.2f}",
            "CPUTIME/WCTIME in percent",
        ),
        "MAXVM": (f"{max_memory_kib:.0f}", "maximum virtual memory used in KiB"),
        "TIMEOUT": (str(timeout).lower(), "did the solver exceed a time limit?"),
        "MEMOUT": (str(memout).lower(), "did the solver exceed the memory limit?"),
    }

    with Path.open(value_file, "w") as f:
        for key, (value, comment) in stats.items():
            f.write(f"# {key}: {comment}\n")
            f.write(f"{key}={value}\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run and monitor a command.")
    parser.add_argument(
        "--timestamp",
        action="store_true",
        help="Include a timestamp in the output logs.",
    )
    parser.add_argument("-C", "--cpu-limit", type=int, help="CPU time limit in seconds.")
    parser.add_argument(
        "-W", "--wall-clock-limit", type=int, help="Wall clock time limit in seconds."
    )
    parser.add_argument(
        "-V",
        "--vm-limit",
        type=int,
        help="Virtual memory limit in KiB.",
    )
    parser.add_argument(
        "-w",
        "--watcher-data",
        type=Path,
        required=True,
        help="File to write watcher info to.",
    )
    parser.add_argument(
        "-v",
        "--var",
        type=Path,
        required=True,
        help="File to write final resource values to.",
    )
    parser.add_argument(
        "-o",
        "--solver-data",
        required=False,
        type=Path,
        help="File to redirect command stdout and stderr to.",
    )
    parser.add_argument("command", nargs=argparse.REMAINDER, help="The command to run.")

    args = parser.parse_args()

    if not args.command:
        parser.error("You must specify a command to run.")

    run_with_monitoring(
        command=args.command,
        watcher_file=args.watcher_data,
        value_file=args.var,
        output_file=args.solver_data,
        timestamp=args.timestamp,
        cpu_limit=args.cpu_limit,
        wall_clock_limit=args.wall_clock_limit,
        vm_limit=args.vm_limit,
    )



"""Classes and Enums to control settings."""

from __future__ import annotations
import configparser
import argparse
from enum import Enum
from pathlib import Path
from typing import Any, NamedTuple, Optional

from runrunner import Runner

from sparkle.types import SparkleObjective, resolve_objective
from sparkle.configurator.configurator import Configurator
from sparkle.configurator import implementations as cim
from sparkle.platform.cli_types import VerbosityLevel


class Option(NamedTuple):
    """Class to define an option in the Settings."""

    name: str
    section: str
    type: Any
    default_value: Any
    alternatives: tuple[str, ...]
    help: str = ""
    cli_kwargs: dict[str, Any] = {}

    def __str__(self: Option) -> str:
        """Return the option name."""
        return self.name

    def __eq__(self: Option, other: Any) -> bool:
        """Check if two options are equal."""
        if isinstance(other, Option):
            return (
                self.name == other.name
                and self.section == other.section
                and self.type == other.type
                and self.default_value == other.default_value
                and self.alternatives == other.alternatives
            )
        if isinstance(other, str):
            return self.name == other or other in self.alternatives
        return False

    @property
    def args(self: Option) -> list[str]:
        """Return the option names as a command line arguments."""
        return [
            f"--{name.replace('_', '-')}"
            for name in [self.name] + list(self.alternatives)
        ]

    @property
    def kwargs(self: Option) -> dict[str, Any]:
        """Return the option attributes as kwargs."""
        kw = {"help": self.help, **self.cli_kwargs}

        # If this option uses a boolean flag action, argparse must NOT receive 'type'
        action = kw.get("action")
        if action in ("store_true", "store_false"):
            return kw

        # Otherwise include the base 'type'
        return {"type": self.type, **kw}


class Settings:
    """Class to read, write, set, and get settings."""

    # CWD Prefix
    cwd_prefix = Path()  # Empty for now

    # Library prefix
    lib_prefix = Path(__file__).parent.parent.resolve()

    # Default directory names
    rawdata_dir = Path("Raw_Data")
    analysis_dir = Path("Analysis")
    DEFAULT_settings_dir = Path("Settings")
    __settings_file = Path("sparkle_settings.ini")
    __latest_settings_file = Path("latest.ini")

    # Default settings path
    DEFAULT_settings_path = Path(cwd_prefix / DEFAULT_settings_dir / __settings_file)
    DEFAULT_previous_settings_path = Path(
        cwd_prefix / DEFAULT_settings_dir / __latest_settings_file
    )
    DEFAULT_reference_dir = DEFAULT_settings_dir / "Reference_Lists"

    # Default library pathing
    DEFAULT_components = lib_prefix / "Components"

    # Report Component: Bilbiography
    bibliography_path = DEFAULT_components / "latex_source" / "report.bib"

    # Example settings path
    DEFAULT_example_settings_path = Path(DEFAULT_components / "sparkle_settings.ini")

    # Runsolver component
    DEFAULT_runsolver_dir = DEFAULT_components / "runsolver" / "src"
    DEFAULT_runsolver_exec = DEFAULT_runsolver_dir / "runsolver"

    # Ablation component
    DEFAULT_ablation_dir = DEFAULT_components / "ablationAnalysis-0.9.4"
    DEFAULT_ablation_exec = DEFAULT_ablation_dir / "ablationAnalysis"
    DEFAULT_ablation_validation_exec = DEFAULT_ablation_dir / "ablationValidation"

    # Default input directory pathing
    DEFAULT_solver_dir = cwd_prefix / "Solvers"
    DEFAULT_instance_dir = cwd_prefix / "Instances"
    DEFAULT_extractor_dir = cwd_prefix / "Extractors"
    DEFAULT_snapshot_dir = cwd_prefix / "Snapshots"

    # Default output directory pathing
    DEFAULT_tmp_output = cwd_prefix / "Tmp"
    DEFAULT_output = cwd_prefix / "Output"
    DEFAULT_configuration_output = DEFAULT_output / "Configuration"
    DEFAULT_selection_output = DEFAULT_output / "Selection"
    DEFAULT_parallel_portfolio_output = DEFAULT_output / "Parallel_Portfolio"
    DEFAULT_ablation_output = DEFAULT_output / "Ablation"
    DEFAULT_log_output = DEFAULT_output / "Log"

    # Default output subdirs
    DEFAULT_output_analysis = DEFAULT_output / analysis_dir

    # Old default output dirs which should be part of something else
    DEFAULT_feature_data = DEFAULT_output / "Feature_Data"
    DEFAULT_performance_data = DEFAULT_output / "Performance_Data"

    # Collection of all working dirs for platform
    DEFAULT_working_dirs = [
        DEFAULT_solver_dir,
        DEFAULT_instance_dir,
        DEFAULT_extractor_dir,
        DEFAULT_output,
        DEFAULT_configuration_output,
        DEFAULT_selection_output,
        DEFAULT_output_analysis,
        DEFAULT_tmp_output,
        DEFAULT_log_output,
        DEFAULT_feature_data,
        DEFAULT_performance_data,
        DEFAULT_settings_dir,
        DEFAULT_reference_dir,
    ]

    # Old default file paths from GV which should be turned into variables
    DEFAULT_feature_data_path = DEFAULT_feature_data / "feature_data.csv"
    DEFAULT_performance_data_path = DEFAULT_performance_data / "performance_data.csv"

    # Define sections and options
    # GENERAL Options
    SECTION_general: str = "general"
    OPTION_objectives = Option(
        "objectives",
        SECTION_general,
        str,
        None,
        ("sparkle_objectives",),
        "A list of Sparkle objectives.",
        cli_kwargs={"nargs": "+"},
    )
    OPTION_configurator = Option(
        "configurator",
        SECTION_general,
        str,
        None,
        tuple(),
        "Name of the configurator to use.",
    )
    OPTION_solver_cutoff_time = Option(
        "solver_cutoff_time",
        SECTION_general,
        int,
        None,
        ("target_cutoff_time", "cutoff_time_each_solver_call"),
        "Solver cutoff time in seconds.",
    )
    OPTION_extractor_cutoff_time = Option(
        "extractor_cutoff_time",
        SECTION_general,
        int,
        None,
        tuple("cutoff_time_each_feature_computation"),
        "Extractor cutoff time in seconds.",
    )
    OPTION_run_on = Option(
        "run_on",
        SECTION_general,
        Runner,
        None,
        tuple(),
        "On which compute resource to execute.",
        cli_kwargs={"choices": [Runner.LOCAL, Runner.SLURM]},
    )
    OPTION_verbosity = Option(
        "verbosity",
        SECTION_general,
        VerbosityLevel,
        VerbosityLevel.STANDARD,
        ("verbosity_level",),
        "Verbosity level.",
    )
    OPTION_seed = Option(
        "seed",
        SECTION_general,
        int,
        None,
        tuple(),
        "Seed to use for pseudo-random number generators.",
    )
    OPTION_appendices = Option(
        "appendices",
        SECTION_general,
        bool,
        False,
        tuple(),
        "Include the appendix section in the generated report.",
        cli_kwargs={
            "action": "store_true",
            "default": None,
        },
    )

    # CONFIGURATION Options
    SECTION_configuration = "configuration"
    OPTION_configurator_number_of_runs = Option(
        "number_of_runs",
        SECTION_configuration,
        int,
        None,
        tuple(),
        "The number of independent configurator jobs/runs.",
    )
    OPTION_configurator_solver_call_budget = Option(
        "solver_calls",
        SECTION_configuration,
        int,
        None,
        tuple(),
        "The maximum number of calls (evaluations) a configurator can do in a single "
        "run of the solver.",
    )
    OPTION_configurator_max_iterations = Option(
        "max_iterations",
        SECTION_configuration,
        int,
        None,
        ("maximum_iterations",),
        "The maximum number of iterations a configurator can do in a single job.",
    )

    # ABLATION Options
    SECTION_ablation = "ablation"
    OPTION_ablation_racing = Option(
        "racing",
        SECTION_ablation,
        bool,
        False,
        ("ablation_racing",),
        "Set a flag indicating whether racing should be used for ablation.",
    )
    OPTION_ablation_clis_per_node = Option(
        "clis_per_node",
        SECTION_ablation,
        int,
        None,
        (
            "max_parallel_runs_per_node",
            "maximum_parallel_runs_per_node",
        ),
        "The maximum number of ablation analysis jobs to run in parallel on a single "
        "node.",
    )

    # SELECTION Options
    SECTION_selection = "selection"
    OPTION_selection_class = Option(
        "selector_class",
        SECTION_selection,
        str,
        None,
        ("class",),
        "Can contain any of the class names as defined in asf.selectors.",
    )
    OPTION_selection_model = Option(
        "selector_model",
        SECTION_selection,
        str,
        None,
        ("model",),
        "Can be any of the sklearn.ensemble models.",
    )
    OPTION_minimum_marginal_contribution = Option(
        "minimum_marginal_contribution",
        SECTION_selection,
        float,
        0.01,
        (
            "minimum_marginal_contribution",
            "minimum_contribution",
            "contribution_threshold",
        ),
        "The minimum marginal contribution a solver (configuration) must have to be used for the selector.",
    )

    # SMAC2 Options
    SECTION_smac2 = "smac2"
    OPTION_smac2_wallclock_time_budget = Option(
        "wallclock_time_budget",
        SECTION_smac2,
        int,
        None,
        ("wallclock_time",),
        "The wallclock time budget in seconds for each SMAC2 run.",
    )
    OPTION_smac2_cpu_time_budget = Option(
        "cpu_time_budget",
        SECTION_smac2,
        int,
        None,
        ("cpu_time",),
        "The cpu time budget in seconds for each SMAC2 run.",
    )
    OPTION_smac2_target_cutoff_length = Option(
        "target_cutoff_length",
        SECTION_smac2,
        str,
        None,
        ("cutoff_length", "solver_cutoff_length"),
        "The target cutoff length for SMAC2 solver call.",
    )
    OPTION_smac2_count_tuner_time = Option(
        "use_cpu_time_in_tunertime",
        SECTION_smac2,
        bool,
        None,
        ("countSMACTimeAsTunerTime",),
        "Whether to count and deducted SMAC2 CPU time from the CPU time budget.",
    )
    OPTION_smac2_cli_cores = Option(
        "cli_cores",
        SECTION_smac2,
        int,
        None,
        tuple(),
        "Number of cores to use to execute SMAC2 runs.",
    )
    OPTION_smac2_max_iterations = Option(
        "max_iterations",
        SECTION_smac2,
        int,
        None,
        (
            "iteration_limit",
            "numIterations",
            "numberOfIterations",
        ),
        "The maximum number of iterations SMAC2 configurator can do in a single job.",
    )

    # SMAC3 Options
    SECTION_smac3 = "smac3"
    OPTION_smac3_number_of_trials = Option(
        "n_trials",
        SECTION_smac3,
        int,
        None,
        ("n_trials", "number_of_trials", "solver_calls"),
        "Maximum calls SMAC3 is allowed to make to the Solver in a single run/job.",
    )
    OPTION_smac3_facade = Option(
        "facade",
        SECTION_smac3,
        str,
        "AlgorithmConfigurationFacade",
        ("facade", "smac_facade", "smac3_facade"),
        "The SMAC3 Facade to use. See the SMAC3 documentation for more options.",
    )
    OPTION_smac3_facade_max_ratio = Option(
        "facade_max_ratio",
        SECTION_smac3,
        float,
        None,
        ("facade_max_ratio", "smac3_facade_max_ratio", "smac3_facade_max_ratio"),
        "The SMAC3 Facade max ratio. See the SMAC3 documentation for more options.",
    )
    OPTION_smac3_crash_cost = Option(
        "crash_cost",
        SECTION_smac3,
        float,
        None,
        tuple(),
        "Defines the cost for a failed trial, defaults in SMAC3 to np.inf.",
    )
    OPTION_smac3_termination_cost_threshold = Option(
        "termination_cost_threshold",
        SECTION_smac3,
        float,
        None,
        tuple(),
        "Defines a cost threshold when the SMAC3 optimization should stop.",
    )
    OPTION_smac3_wallclock_time_budget = Option(
        "walltime_limit",
        SECTION_smac3,
        float,
        None,
        ("wallclock_time", "wallclock_budget", "wallclock_time_budget"),
        "The maximum time in seconds that SMAC3 is allowed to run per job.",
    )
    OPTION_smac3_cpu_time_budget = Option(
        "cputime_limit",
        SECTION_smac3,
        float,
        None,
        ("cpu_time", "cpu_budget", "cpu_time_budget"),
        "The maximum CPU time in seconds that SMAC3 is allowed to run per job.",
    )
    OPTION_smac3_use_default_config = Option(
        "use_default_config",
        SECTION_smac3,
        bool,
        None,
        tuple(),
        "If True, the configspace's default configuration is evaluated in the initial "
        "design. For historic benchmark reasons, this is False by default. Notice, that "
        "this will result in n_configs + 1 for the initial design. Respecting n_trials, "
        "this will result in one fewer evaluated configuration in the optimization.",
    )
    OPTION_smac3_min_budget = Option(
        "min_budget",
        SECTION_smac3,
        float,
        None,
        ("minimum_budget",),
        "The minimum budget (epochs, subset size, number of instances, ...) that is used"
        " for the optimization. Use this argument if you use multi-fidelity or instance "
        "optimization.",
    )
    OPTION_smac3_max_budget = Option(
        "max_budget",
        SECTION_smac3,
        float,
        None,
        ("maximum_budget",),
        "The maximum budget (epochs, subset size, number of instances, ...) that is used"
        " for the optimization. Use this argument if you use multi-fidelity or instance "
        "optimization.",
    )

    # IRACE Options
    SECTION_irace = "irace"
    OPTION_irace_max_time = Option(
        "max_time",
        SECTION_irace,
        int,
        0,
        ("maximum_time",),
        "The maximum time in seconds for each IRACE run/job.",
    )
    OPTION_irace_max_experiments = Option(
        "max_experiments",
        SECTION_irace,
        int,
        0,
        ("maximum_experiments",),
        "The maximum number of experiments for each IRACE run/job.",
    )
    OPTION_irace_first_test = Option(
        "first_test",
        SECTION_irace,
        int,
        None,
        tuple(),
        "Specifies how many instances are evaluated before the first elimination test. "
        "IRACE Default: 5.",
    )
    OPTION_irace_mu = Option(
        "mu",
        SECTION_irace,
        int,
        None,
        tuple(),
        "Parameter used to define the number of configurations sampled and evaluated at "
        "each iteration. IRACE Default: 5.",
    )
    OPTION_irace_max_iterations = Option(
        "max_iterations",
        SECTION_irace,
        int,
        None,
        ("nb_iterations", "iterations", "max_iterations"),
        "Maximum number of iterations to be executed. Each iteration involves the "
        "generation of new configurations and the use of racing to select the best "
        "configurations. By default (with 0), irace calculates a minimum number of "
        "iterations as N^iter = ⌊2 + log2 N param⌋, where N^param is the number of "
        "non-fixed parameters to be tuned. Setting this parameter may make irace stop "
        "sooner than it should without using all the available budget. IRACE recommends"
        " to use the default value (Empty).",
    )

    # ParamILS Options
    SECTION_paramils = "paramils"
    OPTION_paramils_min_runs = Option(
        "min_runs",
        SECTION_paramils,
        int,
        None,
        ("minimum_runs",),
        "Set the minimum number of runs for ParamILS for each run/job.",
    )
    OPTION_paramils_max_runs = Option(
        "max_runs",
        SECTION_paramils,
        int,
        None,
        ("maximum_runs",),
        "Set the maximum number of runs for ParamILS for each run/job.",
    )
    OPTION_paramils_cpu_time_budget = Option(
        "cputime_budget",
        SECTION_paramils,
        int,
        None,
        (
            "cputime_limit",
            "cputime_limit",
            "tunertime_limit",
            "tuner_timeout",
            "tunerTimeout",
        ),
        "The maximum CPU time for each ParamILS run/job.",
    )
    OPTION_paramils_random_restart = Option(
        "random_restart",
        SECTION_paramils,
        float,
        None,
        tuple(),
        "Set the random restart chance for ParamILS.",
    )
    OPTION_paramils_focused = Option(
        "focused_approach",
        SECTION_paramils,
        bool,
        False,
        ("focused",),
        "Set the focused approach for ParamILS.",
    )
    OPTION_paramils_count_tuner_time = Option(
        "use_cpu_time_in_tunertime",
        SECTION_paramils,
        bool,
        None,
        tuple(),
        "Whether to count and deducted ParamILS CPU time from the CPU time budget.",
    )
    OPTION_paramils_cli_cores = Option(
        "cli_cores",
        SECTION_paramils,
        int,
        None,
        tuple(),
        "Number of cores to use for ParamILS runs.",
    )
    OPTION_paramils_max_iterations = Option(
        "max_iterations",
        SECTION_paramils,
        int,
        None,
        (
            "iteration_limit",
            "numIterations",
            "numberOfIterations",
            "maximum_iterations",
        ),
        "The maximum number of ParamILS iterations per run/job.",
    )
    OPTION_paramils_number_initial_configurations = Option(
        "initial_configurations",
        SECTION_paramils,
        int,
        None,
        "The number of initial configurations ParamILS should evaluate.",
    )

    SECTION_parallel_portfolio = "parallel_portfolio"
    OPTION_parallel_portfolio_check_interval = Option(
        "check_interval",
        SECTION_parallel_portfolio,
        int,
        None,
        tuple(),
        "The interval time in seconds when Solvers are checked for their status.",
    )
    OPTION_parallel_portfolio_number_of_seeds_per_solver = Option(
        "num_seeds_per_solver",
        SECTION_parallel_portfolio,
        int,
        None,
        ("solver_seeds",),
        "The number of seeds per solver.",
    )

    SECTION_slurm = "slurm"
    OPTION_slurm_parallel_jobs = Option(
        "number_of_jobs_in_parallel",
        SECTION_slurm,
        int,
        None,
        ("num_job_in_parallel",),
        "The number of jobs to run in parallel.",
    )
    OPTION_slurm_prepend_script = Option(
        "prepend_script",
        SECTION_slurm,
        str,
        None,
        ("job_prepend", "prepend"),
        "Slurm script to prepend to the sbatch.",
    )

    sections_options: dict[str, list[Option]] = {
        SECTION_general: [
            OPTION_objectives,
            OPTION_configurator,
            OPTION_solver_cutoff_time,
            OPTION_extractor_cutoff_time,
            OPTION_run_on,
            OPTION_appendices,
            OPTION_verbosity,
            OPTION_seed,
        ],
        SECTION_configuration: [
            OPTION_configurator_number_of_runs,
            OPTION_configurator_solver_call_budget,
            OPTION_configurator_max_iterations,
        ],
        SECTION_ablation: [
            OPTION_ablation_racing,
            OPTION_ablation_clis_per_node,
        ],
        SECTION_selection: [
            OPTION_selection_class,
            OPTION_selection_model,
            OPTION_minimum_marginal_contribution,
        ],
        SECTION_smac2: [
            OPTION_smac2_wallclock_time_budget,
            OPTION_smac2_cpu_time_budget,
            OPTION_smac2_target_cutoff_length,
            OPTION_smac2_count_tuner_time,
            OPTION_smac2_cli_cores,
            OPTION_smac2_max_iterations,
        ],
        SECTION_smac3: [
            OPTION_smac3_number_of_trials,
            OPTION_smac3_facade,
            OPTION_smac3_facade_max_ratio,
            OPTION_smac3_crash_cost,
            OPTION_smac3_termination_cost_threshold,
            OPTION_smac3_wallclock_time_budget,
            OPTION_smac3_cpu_time_budget,
            OPTION_smac3_use_default_config,
            OPTION_smac3_min_budget,
            OPTION_smac3_max_budget,
        ],
        SECTION_irace: [
            OPTION_irace_max_time,
            OPTION_irace_max_experiments,
            OPTION_irace_first_test,
            OPTION_irace_mu,
            OPTION_irace_max_iterations,
        ],
        SECTION_paramils: [
            OPTION_paramils_min_runs,
            OPTION_paramils_max_runs,
            OPTION_paramils_cpu_time_budget,
            OPTION_paramils_random_restart,
            OPTION_paramils_focused,
            OPTION_paramils_count_tuner_time,
            OPTION_paramils_cli_cores,
            OPTION_paramils_max_iterations,
            OPTION_paramils_number_initial_configurations,
        ],
        SECTION_parallel_portfolio: [
            OPTION_parallel_portfolio_check_interval,
            OPTION_parallel_portfolio_number_of_seeds_per_solver,
        ],
        SECTION_slurm: [OPTION_slurm_parallel_jobs, OPTION_slurm_prepend_script],
    }

    def __init__(
        self: Settings, file_path: Path, argsv: argparse.Namespace = None
    ) -> None:
        """Initialise a settings object.

        Args:
            file_path (Path): Path to the settings file.
            argsv: The CLI arguments to process.
        """
        # Settings 'dictionary' in configparser format
        self.__settings = configparser.ConfigParser()
        for section in self.sections_options.keys():
            self.__settings.add_section(section)
            self.__settings[section] = {}

        # General attributes
        self.__sparkle_objectives: list[SparkleObjective] = None
        self.__general_sparkle_configurator: Configurator = None
        self.__solver_cutoff_time: int = None
        self.__extractor_cutoff_time: int = None
        self.__run_on: Runner = None
        self.__appendices: bool = False
        self.__verbosity_level: VerbosityLevel = None
        self.__seed: Optional[int] = None

        # Configuration attributes
        self.__configurator_solver_call_budget: int = None
        self.__configurator_number_of_runs: int = None
        self.__configurator_max_iterations: int = None

        # Ablation attributes
        self.__ablation_racing_flag: bool = None
        self.__ablation_max_parallel_runs_per_node: int = None

        # Selection attributes
        self.__selection_model: str = None
        self.__selection_class: str = None
        self.__minimum_marginal_contribution: float = None

        # SMAC2 attributes
        self.__smac2_wallclock_time_budget: int = None
        self.__smac2_cpu_time_budget: int = None
        self.__smac2_target_cutoff_length: str = None
        self.__smac2_use_tunertime_in_cpu_time_budget: bool = None
        self.__smac2_cli_cores: int = None
        self.__smac2_max_iterations: int = None

        # SMAC3 attributes
        self.__smac3_number_of_trials: int = None
        self.__smac3_facade: str = None
        self.__smac3_facade_max_ratio: float = None
        self.__smac3_crash_cost: float = None
        self.__smac3_termination_cost_threshold: float = None
        self.__smac3_wallclock_time_limit: int = None
        self.__smac3_cputime_limit: int = None
        self.__smac3_use_default_config: bool = None
        self.__smac3_min_budget: float = None
        self.__smac3_max_budget: float = None

        # IRACE attributes
        self.__irace_max_time: int = None
        self.__irace_max_experiments: int = None
        self.__irace_first_test: int = None
        self.__irace_mu: int = None
        self.__irace_max_iterations: int = None

        # ParamILS attributes
        self.__paramils_cpu_time_budget: int = None
        self.__paramils_min_runs: int = None
        self.__paramils_max_runs: int = None
        self.__paramils_random_restart: float = None
        self.__paramils_focused_approach: bool = None
        self.__paramils_use_cpu_time_in_tunertime: bool = None
        self.__paramils_cli_cores: int = None
        self.__paramils_max_iterations: int = None
        self.__paramils_number_initial_configurations: int = None

        # Parallel portfolio attributes
        self.__parallel_portfolio_check_interval: int = None
        self.__parallel_portfolio_num_seeds_per_solver: int = None

        # Slurm attributes
        self.__slurm_jobs_in_parallel: int = None
        self.__slurm_job_prepend: str = None

        # The seed that has been used to set the random state
        self.random_state: Optional[int] = None

        if file_path and file_path.exists():
            self.read_settings_ini(file_path)

        if argsv:
            self.apply_arguments(argsv)

    def read_settings_ini(self: Settings, file_path: Path) -> None:
        """Read the settings from an INI file."""
        if not file_path.exists():
            raise ValueError(f"Settings file {file_path} does not exist.")
        # Read file
        file_settings = configparser.ConfigParser()
        file_settings.read(file_path)

        # Set internal settings based on data read from FILE if they were read
        # successfully
        if file_settings.sections() == []:
            # Print error if unable to read the settings
            print(
                f"ERROR: Failed to read settings from {file_path} The file may "
                "have been empty or be in another format than INI."
            )
            return

        for section in file_settings.sections():
            if section not in self.__settings.sections():
                print(f'Unrecognised section: "{section}" in file {file_path} ignored')
                continue
            for option_name in file_settings.options(section):
                if option_name not in self.sections_options[section]:
                    if section == Settings.SECTION_slurm:  # Flexible section
                        self.__settings.set(
                            section,
                            option_name,
                            file_settings.get(section, option_name),
                        )
                    else:
                        print(
                            f'Unrecognised section - option combination: "{section} '
                            f'{option_name}" in file {file_path} ignored'
                        )
                    continue
                option_index = self.sections_options[section].index(option_name)
                option = self.sections_options[section][option_index]
                self.__settings.set(
                    section, option.name, file_settings.get(section, option_name)
                )
        del file_settings

    def write_settings_ini(self: Settings, file_path: Path) -> None:
        """Write the settings to an INI file."""
        # Create needed directories if they don't exist
        file_path.parent.mkdir(parents=True, exist_ok=True)
        # We don't write empty sections
        for s in self.__settings.sections():
            if not self.__settings[s]:
                self.__settings.remove_section(s)
        with file_path.open("w") as fout:
            self.__settings.write(fout)
        for s in self.sections_options.keys():
            if s not in self.__settings.sections():
                self.__settings.add_section(s)

    def write_used_settings(self: Settings) -> None:
        """Write the used settings to the default locations."""
        # Write to latest settings file
        self.write_settings_ini(self.DEFAULT_previous_settings_path)

    def apply_arguments(self: Settings, argsv: argparse.Namespace) -> None:
        """Apply the arguments to the settings."""
        # Read a possible second file, that overwrites the first, where applicable
        # e.g. settings are not deleted, but overwritten where applicable
        if hasattr(argsv, "settings_file") and argsv.settings_file:
            self.read_settings_ini(argsv.settings_file)
        # Match all possible arguments to the settings
        for argument in argsv.__dict__.keys():
            value = argsv.__dict__[argument]
            if value is None:
                continue  # Skip None
            value = value.name if isinstance(value, Enum) else str(value)
            for section in self.sections_options.keys():
                if argument in self.sections_options[section]:
                    index = self.sections_options[section].index(argument)
                    option = self.sections_options[section][index]
                    self.__settings.set(option.section, option.name, value)
                    break

    def _abstract_getter(self: Settings, option: Option) -> Any:
        """Abstract getter method."""
        if self.__settings.has_option(option.section, option.name):
            if option.type is bool:
                return self.__settings.getboolean(option.section, option.name)
            value = self.__settings.get(option.section, option.name)
            if not isinstance(value, option.type):
                if issubclass(option.type, Enum):
                    return option.type[value.upper()]
                return option.type(value)  # Attempt to resolve str to type
            return value
        return option.default_value

    # General settings ###
    @property
    def objectives(self: Settings) -> list[SparkleObjective]:
        """Get the objectives for Sparkle."""
        if self.__sparkle_objectives is None and self.__settings.has_option(
            Settings.SECTION_general, "objectives"
        ):
            objectives = self.__settings[Settings.SECTION_general]["objectives"]
            if "status" not in objectives:
                objectives += ",status:metric"
            if "cpu_time" not in objectives:
                objectives += ",cpu_time:metric"
            if "wall_time" not in objectives:
                objectives += ",wall_time:metric"
            if "memory" not in objectives:
                objectives += ",memory:metric"
            self.__sparkle_objectives = [
                resolve_objective(obj) for obj in objectives.split(",")
            ]
        return self.__sparkle_objectives

    @property
    def configurator(self: Settings) -> Configurator:
        """Get the configurator class (instance)."""
        if self.__general_sparkle_configurator is None and self.__settings.has_option(
            Settings.OPTION_configurator.section, Settings.OPTION_configurator.name
        ):
            self.__general_sparkle_configurator = cim.resolve_configurator(
                self.__settings.get(
                    Settings.OPTION_configurator.section,
                    Settings.OPTION_configurator.name,
                )
            )()
        return self.__general_sparkle_configurator

    @property
    def solver_cutoff_time(self: Settings) -> int:
        """Solver cutoff time in seconds."""
        if self.__solver_cutoff_time is None:
            self.__solver_cutoff_time = self._abstract_getter(
                Settings.OPTION_solver_cutoff_time
            )
        return self.__solver_cutoff_time

    @property
    def extractor_cutoff_time(self: Settings) -> int:
        """Extractor cutoff time in seconds."""
        if self.__extractor_cutoff_time is None:
            self.__extractor_cutoff_time = self._abstract_getter(
                Settings.OPTION_extractor_cutoff_time
            )
        return self.__extractor_cutoff_time

    @property
    def run_on(self: Settings) -> Runner:
        """On which compute to run (Local or Slurm)."""
        if self.__run_on is None:
            self.__run_on = self._abstract_getter(Settings.OPTION_run_on)
        return self.__run_on

    @property
    def appendices(self: Settings) -> bool:
        """Whether to include appendices in the report."""
        return self._abstract_getter(Settings.OPTION_appendices)

    @property
    def verbosity_level(self: Settings) -> VerbosityLevel:
        """Verbosity level to use in CLI commands."""
        if self.__verbosity_level is None:
            if self.__settings.has_option(
                Settings.OPTION_verbosity.section, Settings.OPTION_verbosity.name
            ):
                self.__verbosity_level = VerbosityLevel[
                    self.__settings.get(
                        Settings.OPTION_verbosity.section,
                        Settings.OPTION_verbosity.name,
                    )
                ]
            else:
                self.__verbosity_level = Settings.OPTION_verbosity.default_value
        return self.__verbosity_level

    @property
    def seed(self: Settings) -> int:
        """Seed to use in CLI commands."""
        if self.__seed is not None:
            return self.__seed

        section, name = Settings.OPTION_seed.section, Settings.OPTION_seed.name
        if self.__settings.has_option(section, name):
            value = self.__settings.get(section, name)
            self.__seed = int(value)
        else:
            self.__seed = Settings.OPTION_seed.default_value

        return self.__seed

    @seed.setter
    def seed(self: Settings, value: int) -> None:
        """Set the seed value (overwrites settings)."""
        self.__seed = value
        self.__settings.set(
            Settings.OPTION_seed.section, Settings.OPTION_seed.name, str(self.__seed)
        )

    # Configuration settings ###
    @property
    def configurator_solver_call_budget(self: Settings) -> int:
        """The amount of calls a configurator can do to the solver."""
        if self.__configurator_solver_call_budget is None:
            self.__configurator_solver_call_budget = self._abstract_getter(
                Settings.OPTION_configurator_solver_call_budget
            )
        return self.__configurator_solver_call_budget

    @property
    def configurator_number_of_runs(self: Settings) -> int:
        """Get the amount of configurator runs to do."""
        if self.__configurator_number_of_runs is None:
            self.__configurator_number_of_runs = self._abstract_getter(
                Settings.OPTION_configurator_number_of_runs
            )
        return self.__configurator_number_of_runs

    @property
    def configurator_max_iterations(self: Settings) -> int:
        """Get the amount of configurator iterations to do."""
        if self.__configurator_max_iterations is None:
            self.__configurator_max_iterations = self._abstract_getter(
                Settings.OPTION_configurator_max_iterations
            )
        return self.__configurator_max_iterations

    # Ablation settings ###
    @property
    def ablation_racing_flag(self: Settings) -> bool:
        """Get the ablation racing flag."""
        if self.__ablation_racing_flag is None:
            self.__ablation_racing_flag = self._abstract_getter(
                Settings.OPTION_ablation_racing
            )
        return self.__ablation_racing_flag

    @property
    def ablation_max_parallel_runs_per_node(self: Settings) -> int:
        """Get the ablation max parallel runs per node."""
        if self.__ablation_max_parallel_runs_per_node is None:
            self.__ablation_max_parallel_runs_per_node = self._abstract_getter(
                Settings.OPTION_ablation_clis_per_node
            )
        return self.__ablation_max_parallel_runs_per_node

    # Selection settings ###
    @property
    def selection_model(self: Settings) -> str:
        """Get the selection model."""
        if self.__selection_model is None:
            self.__selection_model = self._abstract_getter(
                Settings.OPTION_selection_model
            )
        return self.__selection_model

    @property
    def selection_class(self: Settings) -> str:
        """Get the selection class."""
        if self.__selection_class is None:
            self.__selection_class = self._abstract_getter(
                Settings.OPTION_selection_class
            )
        return self.__selection_class

    @property
    def minimum_marginal_contribution(self: Settings) -> float:
        """Get the minimum marginal contribution."""
        if self.__minimum_marginal_contribution is None:
            self.__minimum_marginal_contribution = self._abstract_getter(
                Settings.OPTION_minimum_marginal_contribution
            )
        return self.__minimum_marginal_contribution

    # Configuration: SMAC2 specific settings ###
    @property
    def smac2_wallclock_time_budget(self: Settings) -> int:
        """Return the SMAC2 wallclock budget per configuration run in seconds."""
        if self.__smac2_wallclock_time_budget is None:
            self.__smac2_wallclock_time_budget = self._abstract_getter(
                Settings.OPTION_smac2_wallclock_time_budget
            )
        return self.__smac2_wallclock_time_budget

    @property
    def smac2_cpu_time_budget(self: Settings) -> int:
        """Return the SMAC2 CPU budget per configuration run in seconds."""
        if self.__smac2_cpu_time_budget is None:
            self.__smac2_cpu_time_budget = self._abstract_getter(
                Settings.OPTION_smac2_cpu_time_budget
            )
        return self.__smac2_cpu_time_budget

    @property
    def smac2_target_cutoff_length(self: Settings) -> str:
        """Return the SMAC2 target cutoff length."""
        if self.__smac2_target_cutoff_length is None:
            self.__smac2_target_cutoff_length = self._abstract_getter(
                Settings.OPTION_smac2_target_cutoff_length
            )
        return self.__smac2_target_cutoff_length

    @property
    def smac2_use_tunertime_in_cpu_time_budget(self: Settings) -> bool:
        """Return whether SMAC2 time should be used in CPU time budget."""
        if self.__smac2_use_tunertime_in_cpu_time_budget is None:
            self.__smac2_use_tunertime_in_cpu_time_budget = self._abstract_getter(
                Settings.OPTION_smac2_count_tuner_time
            )
        return self.__smac2_use_tunertime_in_cpu_time_budget

    @property
    def smac2_cli_cores(self: Settings) -> int:
        """Return the SMAC2 CLI cores."""
        if self.__smac2_cli_cores is None:
            self.__smac2_cli_cores = self._abstract_getter(
                Settings.OPTION_smac2_cli_cores
            )
        return self.__smac2_cli_cores

    @property
    def smac2_max_iterations(self: Settings) -> int:
        """Return the SMAC2 max iterations."""
        if self.__smac2_max_iterations is None:
            self.__smac2_max_iterations = self._abstract_getter(
                Settings.OPTION_smac2_max_iterations
            )
        return self.__smac2_max_iterations

    # SMAC3 attributes ###
    @property
    def smac3_number_of_trials(self: Settings) -> int:
        """Return the SMAC3 number of trials."""
        if self.__smac3_number_of_trials is None:
            self.__smac3_number_of_trials = self._abstract_getter(
                Settings.OPTION_smac3_number_of_trials
            )
        return self.__smac3_number_of_trials

    @property
    def smac3_facade(self: Settings) -> str:
        """Return the SMAC3 facade."""
        if self.__smac3_facade is None:
            self.__smac3_facade = self._abstract_getter(Settings.OPTION_smac3_facade)
        return self.__smac3_facade

    @property
    def smac3_facade_max_ratio(self: Settings) -> float:
        """Return the SMAC3 facade max ratio."""
        if self.__smac3_facade_max_ratio is None:
            self.__smac3_facade_max_ratio = self._abstract_getter(
                Settings.OPTION_smac3_facade_max_ratio
            )
        return self.__smac3_facade_max_ratio

    @property
    def smac3_crash_cost(self: Settings) -> float:
        """Return the SMAC3 crash cost."""
        if self.__smac3_crash_cost is None:
            self.__smac3_crash_cost = self._abstract_getter(
                Settings.OPTION_smac3_crash_cost
            )
        return self.__smac3_crash_cost

    @property
    def smac3_termination_cost_threshold(self: Settings) -> float:
        """Return the SMAC3 termination cost threshold."""
        if self.__smac3_termination_cost_threshold is None:
            self.__smac3_termination_cost_threshold = self._abstract_getter(
                Settings.OPTION_smac3_termination_cost_threshold
            )
        return self.__smac3_termination_cost_threshold

    @property
    def smac3_wallclock_time_budget(self: Settings) -> int:
        """Return the SMAC3 walltime budget in seconds."""
        if self.__smac3_wallclock_time_limit is None:
            self.__smac3_wallclock_time_limit = self._abstract_getter(
                Settings.OPTION_smac3_wallclock_time_budget
            )
        return self.__smac3_wallclock_time_limit

    @property
    def smac3_cpu_time_budget(self: Settings) -> int:
        """Return the SMAC3 cputime budget in seconds."""
        if self.__smac3_cputime_limit is None:
            self.__smac3_cputime_limit = self._abstract_getter(
                Settings.OPTION_smac3_cpu_time_budget
            )
        return self.__smac3_cputime_limit

    @property
    def smac3_use_default_config(self: Settings) -> bool:
        """Return whether SMAC3 should use the default config."""
        if self.__smac3_use_default_config is None:
            self.__smac3_use_default_config = self._abstract_getter(
                Settings.OPTION_smac3_use_default_config
            )
        return self.__smac3_use_default_config

    @property
    def smac3_min_budget(self: Settings) -> int:
        """Return the SMAC3 min budget."""
        if self.__smac3_min_budget is None:
            self.__smac3_min_budget = self._abstract_getter(
                Settings.OPTION_smac3_min_budget
            )
        return self.__smac3_min_budget

    @property
    def smac3_max_budget(self: Settings) -> int:
        """Return the SMAC3 max budget."""
        if self.__smac3_max_budget is None:
            self.__smac3_max_budget = self._abstract_getter(
                Settings.OPTION_smac3_max_budget
            )
        return self.__smac3_max_budget

    # IRACE settings ###
    @property
    def irace_max_time(self: Settings) -> int:
        """Return the max time in seconds for IRACE."""
        if self.__irace_max_time is None:
            self.__irace_max_time = self._abstract_getter(Settings.OPTION_irace_max_time)
        return self.__irace_max_time

    @property
    def irace_max_experiments(self: Settings) -> int:
        """Return the max experiments for IRACE."""
        if self.__irace_max_experiments is None:
            self.__irace_max_experiments = self._abstract_getter(
                Settings.OPTION_irace_max_experiments
            )
        return self.__irace_max_experiments

    @property
    def irace_first_test(self: Settings) -> int:
        """Return the first test for IRACE."""
        if self.__irace_first_test is None:
            self.__irace_first_test = self._abstract_getter(
                Settings.OPTION_irace_first_test
            )
        return self.__irace_first_test

    @property
    def irace_mu(self: Settings) -> int:
        """Return the mu for IRACE."""
        if self.__irace_mu is None:
            self.__irace_mu = self._abstract_getter(Settings.OPTION_irace_mu)
        return self.__irace_mu

    @property
    def irace_max_iterations(self: Settings) -> int:
        """Return the max iterations for IRACE."""
        if self.__irace_max_iterations is None:
            self.__irace_max_iterations = self._abstract_getter(
                Settings.OPTION_irace_max_iterations
            )
        return self.__irace_max_iterations

    # ParamILS settings ###
    @property
    def paramils_cpu_time_budget(self: Settings) -> int:
        """Return the CPU time budget for ParamILS."""
        if self.__paramils_cpu_time_budget is None:
            self.__paramils_cpu_time_budget = self._abstract_getter(
                Settings.OPTION_paramils_cpu_time_budget
            )
        return self.__paramils_cpu_time_budget

    @property
    def paramils_min_runs(self: Settings) -> int:
        """Return the min runs for ParamILS."""
        if self.__paramils_min_runs is None:
            self.__paramils_min_runs = self._abstract_getter(
                Settings.OPTION_paramils_min_runs
            )
        return self.__paramils_min_runs

    @property
    def paramils_max_runs(self: Settings) -> int:
        """Return the max runs for ParamILS."""
        if self.__paramils_max_runs is None:
            self.__paramils_max_runs = self._abstract_getter(
                Settings.OPTION_paramils_max_runs
            )
        return self.__paramils_max_runs

    @property
    def paramils_random_restart(self: Settings) -> float:
        """Return the random restart for ParamILS."""
        if self.__paramils_random_restart is None:
            self.__paramils_random_restart = self._abstract_getter(
                Settings.OPTION_paramils_random_restart
            )
        return self.__paramils_random_restart

    @property
    def paramils_focused_approach(self: Settings) -> bool:
        """Return the focused approach for ParamILS."""
        if self.__paramils_focused_approach is None:
            self.__paramils_focused_approach = self._abstract_getter(
                Settings.OPTION_paramils_focused
            )
        return self.__paramils_focused_approach

    @property
    def paramils_use_cpu_time_in_tunertime(self: Settings) -> bool:
        """Return the use cpu time for ParamILS."""
        if self.__paramils_use_cpu_time_in_tunertime is None:
            self.__paramils_use_cpu_time_in_tunertime = self._abstract_getter(
                Settings.OPTION_paramils_count_tuner_time
            )
        return self.__paramils_use_cpu_time_in_tunertime

    @property
    def paramils_cli_cores(self: Settings) -> int:
        """The number of CPU cores to use for ParamILS."""
        if self.__paramils_cli_cores is None:
            self.__paramils_cli_cores = self._abstract_getter(
                Settings.OPTION_paramils_cli_cores
            )
        return self.__paramils_cli_cores

    @property
    def paramils_max_iterations(self: Settings) -> int:
        """Return the max iterations for ParamILS."""
        if self.__paramils_max_iterations is None:
            self.__paramils_max_iterations = self._abstract_getter(
                Settings.OPTION_paramils_max_iterations
            )
        return self.__paramils_max_iterations

    @property
    def paramils_number_initial_configurations(self: Settings) -> int:
        """Return the number of initial configurations for ParamILS."""
        if self.__paramils_number_initial_configurations is None:
            self.__paramils_number_initial_configurations = self._abstract_getter(
                Settings.OPTION_paramils_number_initial_configurations
            )
        return self.__paramils_number_initial_configurations

    # Parallel Portfolio settings ###
    @property
    def parallel_portfolio_check_interval(self: Settings) -> int:
        """Return the check interval for the parallel portfolio."""
        if self.__parallel_portfolio_check_interval is None:
            self.__parallel_portfolio_check_interval = self._abstract_getter(
                Settings.OPTION_parallel_portfolio_check_interval
            )
        return self.__parallel_portfolio_check_interval

    @property
    def parallel_portfolio_num_seeds_per_solver(self: Settings) -> int:
        """Return the number of seeds per solver for the parallel portfolio."""
        if self.__parallel_portfolio_num_seeds_per_solver is None:
            self.__parallel_portfolio_num_seeds_per_solver = self._abstract_getter(
                Settings.OPTION_parallel_portfolio_number_of_seeds_per_solver
            )
        return self.__parallel_portfolio_num_seeds_per_solver

    # Slurm settings ###
    @property
    def slurm_jobs_in_parallel(self: Settings) -> int:
        """Return the (maximum) number of jobs to run in parallel."""
        if self.__slurm_jobs_in_parallel is None:
            self.__slurm_jobs_in_parallel = self._abstract_getter(
                Settings.OPTION_slurm_parallel_jobs
            )
        return self.__slurm_jobs_in_parallel

    @property
    def slurm_job_prepend(self: Settings) -> str:
        """Return the slurm job prepend."""
        if self.__slurm_job_prepend is None and self.__settings.has_option(
            Settings.OPTION_slurm_prepend_script.section,
            Settings.OPTION_slurm_prepend_script.name,
        ):
            value = self.__settings[Settings.OPTION_slurm_prepend_script.section][
                Settings.OPTION_slurm_prepend_script.name
            ]
            try:
                path = Path(value)
                if path.is_file():
                    with path.open() as f:
                        value = f.read()
                        f.close()
                self.__slurm_job_prepend = str(value)
            except TypeError:
                pass
        return self.__slurm_job_prepend

    @property
    def sbatch_settings(self: Settings) -> list[str]:
        """Return the sbatch settings."""
        sbatch_options = self.__settings[Settings.SECTION_slurm]
        # Return all non-predefined keys
        return [
            f"--{key}={sbatch_options[key]}"
            for key in sbatch_options.keys()
            if key not in Settings.sections_options[Settings.SECTION_slurm]
        ]

    # General functionalities ###

    def get_configurator_output_path(self: Settings, configurator: Configurator) -> Path:
        """Return the configurator output path."""
        return self.DEFAULT_configuration_output / configurator.name

    def get_configurator_settings(
        self: Settings, configurator_name: str
    ) -> dict[str, any]:
        """Return the settings of a specific configurator."""
        configurator_settings = {
            "solver_calls": self.configurator_solver_call_budget,
            "solver_cutoff_time": self.solver_cutoff_time,
            "max_iterations": self.configurator_max_iterations,
        }
        # In the settings below, we default to the configurator general settings if no
        # specific configurator settings are given, by using the [None] or [Value]
        if configurator_name == cim.SMAC2.__name__:
            # Return all settings from the SMAC2 section
            configurator_settings.update(
                {
                    "cpu_time": self.smac2_cpu_time_budget,
                    "wallclock_time": self.smac2_wallclock_time_budget,
                    "target_cutoff_length": self.smac2_target_cutoff_length,
                    "use_cpu_time_in_tunertime": self.smac2_use_tunertime_in_cpu_time_budget,
                    "cli_cores": self.smac2_cli_cores,
                    "max_iterations": self.smac2_max_iterations
                    or configurator_settings["max_iterations"],
                }
            )
        elif configurator_name == cim.SMAC3.__name__:
            # Return all settings from the SMAC3 section
            del configurator_settings["max_iterations"]  # SMAC3 does not have this?
            configurator_settings.update(
                {
                    "smac_facade": self.smac3_facade,
                    "max_ratio": self.smac3_facade_max_ratio,
                    "crash_cost": self.smac3_crash_cost,
                    "termination_cost_threshold": self.smac3_termination_cost_threshold,
                    "walltime_limit": self.smac3_wallclock_time_budget,
                    "cputime_limit": self.smac3_cpu_time_budget,
                    "use_default_config": self.smac3_use_default_config,
                    "min_budget": self.smac3_min_budget,
                    "max_budget": self.smac3_max_budget,
                    "solver_calls": self.smac3_number_of_trials
                    or configurator_settings["solver_calls"],
                }
            )
            # Do not pass None values to SMAC3, its Scenario resolves default settings
            configurator_settings = {
                key: value
                for key, value in configurator_settings.items()
                if value is not None
            }
        elif configurator_name == cim.IRACE.__name__:
            # Return all settings from the IRACE section
            configurator_settings.update(
                {
                    "solver_calls": self.irace_max_experiments,
                    "max_time": self.irace_max_time,
                    "first_test": self.irace_first_test,
                    "mu": self.irace_mu,
                    "max_iterations": self.irace_max_iterations
                    or configurator_settings["max_iterations"],
                }
            )
            if (
                configurator_settings["solver_calls"] == 0
                and configurator_settings["max_time"] == 0
            ):  # Default to base
                configurator_settings["solver_calls"] = (
                    self.configurator_solver_call_budget
                )
        elif configurator_name == cim.ParamILS.__name__:
            configurator_settings.update(
                {
                    "tuner_timeout": self.paramils_cpu_time_budget,
                    "min_runs": self.paramils_min_runs,
                    "max_runs": self.paramils_max_runs,
                    "focused_ils": self.paramils_focused_approach,
                    "initial_configurations": self.paramils_number_initial_configurations,
                    "random_restart": self.paramils_random_restart,
                    "cli_cores": self.paramils_cli_cores,
                    "use_cpu_time_in_tunertime": self.paramils_use_cpu_time_in_tunertime,
                    "max_iterations": self.paramils_max_iterations
                    or configurator_settings["max_iterations"],
                }
            )
        return configurator_settings

    @staticmethod
    def check_settings_changes(
        cur_settings: Settings, prev_settings: Settings, verbose: bool = True
    ) -> bool:
        """Check if there are changes between the previous and the current settings.

        Prints any section changes, printing None if no setting was found.

        Args:
          cur_settings: The current settings
          prev_settings: The previous settings
          verbose: Verbosity of the function

        Returns:
          True iff there are changes.
        """
        cur_dict = cur_settings.__settings._sections
        prev_dict = prev_settings.__settings._sections

        cur_sections_set = set(cur_dict.keys())
        prev_sections_set = set(prev_dict.keys())

        sections_remained = cur_sections_set & prev_sections_set
        option_changed = False
        for section in sections_remained:
            printed_section = False
            names = set(cur_dict[section].keys()) | set(prev_dict[section].keys())
            if (
                section == "general" and "seed" in names
            ):  # Do not report on the seed, is supposed to change
                names.remove("seed")
            for name in names:
                # if name is not present in one of the two dicts, get None as placeholder
                cur_val = cur_dict[section].get(name, None)
                prev_val = prev_dict[section].get(name, None)

                # If cur val is None, it is default
                if cur_val is not None and cur_val != prev_val:
                    if not option_changed and verbose:  # Print the initial
                        print("[INFO] The following attributes/options have changed:")
                        option_changed = True

                    # do we have yet to print the section?
                    if not printed_section and verbose:
                        print(f"  - In the section '{section}':")
                        printed_section = True

                    # print actual change
                    if verbose:
                        print(f"    · '{name}' changed from '{prev_val}' to '{cur_val}'")

        return option_changed



"""File to handle a Selector for selecting Solvers."""

from __future__ import annotations
import random
from pathlib import Path

from sklearn.base import ClassifierMixin, RegressorMixin
from asf.cli import cli_train as asf_cli
from asf.predictors import AbstractPredictor
from asf.selectors.abstract_model_based_selector import AbstractModelBasedSelector

import runrunner as rrr
from runrunner import Runner, Run
import pandas as pd

from sparkle.types import SparkleObjective, resolve_objective
from sparkle.structures import FeatureDataFrame, PerformanceDataFrame
from sparkle.instance import InstanceSet


class Selector:
    """The Selector class for handling Algorithm Selection."""

    selector_cli = Path(__file__).parent / "selector_cli.py"

    def __init__(
        self: Selector,
        selector_class: AbstractModelBasedSelector,
        model_class: AbstractPredictor | ClassifierMixin | RegressorMixin,
    ) -> None:
        """Initialize the Selector object.

        Args:
            selector_class: The (name of) Selector class to construct.
            model_class: The (name of) model class the selector will use.
        """
        if isinstance(selector_class, str):  # Resolve class name
            from asf import selectors

            selector_class = getattr(selectors, selector_class)
        if isinstance(model_class, str):  # Resolve class name
            from sklearn import ensemble

            model_class = getattr(ensemble, model_class)
        self.selector_class = selector_class
        self.model_class = model_class

    @property
    def name(self: Selector) -> str:
        """Return the name of the selector."""
        return f"{self.selector_class.__name__}_{self.model_class.__name__}"

    def construct(
        self: Selector,
        selection_scenario: SelectionScenario,
        run_on: Runner = Runner.SLURM,
        job_name: str = None,
        sbatch_options: list[str] = None,
        slurm_prepend: str | list[str] | Path = None,
        base_dir: Path = Path(),
    ) -> Run:
        """Construct the Selector.

        Args:
            selection_scenario: The scenario to construct the Selector for.
            run_on: Which runner to use. Defaults to slurm.
            job_name: Name to give the construction job when submitting.
            sbatch_options: Additional options to pass to sbatch.
            slurm_prepend: Slurm script to prepend to the sbatch
            base_dir: The base directory to run the Selector in.

        Returns:
            The construction Run
        """
        selection_scenario.create_scenario()
        selector = self.selector_class(
            model_class=self.model_class,
            budget=selection_scenario.solver_cutoff,
            maximize=not selection_scenario.objective.minimise,
        )
        cmd = asf_cli.build_cli_command(
            selector,
            selection_scenario.feature_target_path,
            selection_scenario.performance_target_path,
            selection_scenario.selector_file_path,
        )
        cmd = [" ".join([str(c) for c in cmd])]

        job_name = job_name or f"Selector Construction: {selection_scenario.name}"
        construct = rrr.add_to_queue(
            runner=run_on,
            cmd=cmd,
            name=job_name,
            base_dir=base_dir,
            sbatch_options=sbatch_options,
            prepend=slurm_prepend,
        )

        if run_on == Runner.LOCAL:
            construct.wait()
            if not selection_scenario.selector_file_path.is_file():
                print(f"Selector construction of {self.name} failed!")
        return construct

    def run(
        self: Selector,
        selector_path: Path,
        instance: str,
        feature_data: FeatureDataFrame,
    ) -> list:
        """Run the Selector, returning the prediction schedule upon success."""
        instance_features = feature_data[
            [
                instance,
            ]
        ]
        instance_features.index = instance_features.index.map("_".join)  # Reduce
        instance_features = instance_features.T  # ASF dataframe structure
        selector = self.selector_class.load(selector_path)
        schedule = selector.predict(instance_features)
        if schedule is None:
            print(f"ERROR: Selector {self.name} failed predict schedule!")
            return None
        # ASF presents result as schedule per instance, we only use one in this setting
        schedule = schedule[instance]
        for index, (solver, time) in enumerate(schedule):
            # Split solver name back into solver and config id
            solver_name, conf_index = solver.split("_", maxsplit=1)
            schedule[index] = (solver_name, conf_index, time)
        return schedule

    def run_cli(
        self: Selector,
        scenario_path: Path,
        instance_set: InstanceSet | list[Path],
        feature_data: Path,
        run_on: Runner = Runner.LOCAL,
        sbatch_options: list[str] = None,
        slurm_prepend: str | list[str] | Path = None,
        job_name: str = None,
        dependencies: list[Run] = None,
        log_dir: Path = None,
    ) -> Run:
        """Run the Selector CLI and write result to the Scenario PerformanceDataFrame.

        Args:
            scenario_path: The path to the scenario with the Selector to run.
            instance_set: The instance set to run the Selector on.
            feature_data: The instance feature data to use.
            run_on: Which runner to use. Defaults to slurm.
            sbatch_options: Additional options to pass to sbatch.
            slurm_prepend: Slurm script to prepend to the sbatch
            job_name: Name to give the Slurm job when submitting.
            dependencies: List of dependencies to add to the job.
            log_dir: The directory to write logs to.

        Returns:
            The Run object.
        """
        # NOTE: The selector object and the scenario selector could differ which could
        # cause unintended behaviour (e.g. running a different selector than desired)
        instances = (
            instance_set
            if isinstance(instance_set, list)
            else instance_set.instance_paths
        )
        commands = [
            f"python3 {Selector.selector_cli} "
            f"--selector-scenario {scenario_path} "
            f"--instance {instance_path} "
            f"--feature-data {feature_data} "
            f"--log-dir {log_dir} "
            f"--seed {random.randint(0, 2**32 - 1)}"
            for instance_path in instances
        ]

        job_name = (
            f"Run Selector: {self.name} on {len(instances)} instances"
            if not job_name
            else job_name
        )
        import subprocess

        r = rrr.add_to_queue(
            cmd=commands,
            name=job_name,
            stdout=None if run_on == Runner.LOCAL else subprocess.PIPE,  # Print
            stderr=None if run_on == Runner.LOCAL else subprocess.PIPE,  # Print
            base_dir=log_dir,
            runner=run_on,
            sbatch_options=sbatch_options,
            prepend=slurm_prepend,
            dependencies=dependencies,
        )
        if run_on == Runner.LOCAL:
            r.wait()
        return r


class SelectionScenario:
    """A scenario for a Selector."""

    __selector_solver_name__ = "portfolio_selector"

    def __init__(
        self: SelectionScenario,
        parent_directory: Path,
        selector: Selector,
        objective: SparkleObjective,
        performance_data: PerformanceDataFrame | Path,
        feature_data: FeatureDataFrame | Path,
        feature_extractors: list[str] = None,
        solver_cutoff: int | float = None,
        extractor_cutoff: int | float = None,
        ablate: bool = False,
        subdir_path: Path = None,
    ) -> None:
        """Initialize a scenario for a selector."""
        self.selector: Selector = selector
        self.objective: SparkleObjective = objective
        self.solver_cutoff: float = solver_cutoff
        self.extractor_cutoff: float = extractor_cutoff
        if subdir_path is not None:
            self.directory = parent_directory / subdir_path
        elif isinstance(performance_data, PerformanceDataFrame):
            self.directory: Path = (
                parent_directory
                / selector.name
                / "_".join([Path(s).name for s in performance_data.solvers])
            )
        else:
            self.directory = performance_data.parent
        self.name = f"{selector.name} on {self.directory.name}"
        self.selector_file_path: Path = self.directory / "portfolio_selector"
        self.scenario_file: Path = self.directory / "scenario.txt"
        self.selector_performance_path: Path = (
            self.directory / "selector_performance.csv"
        )
        if self.selector_performance_path.exists():
            self.selector_performance_data = PerformanceDataFrame(
                self.selector_performance_path
            )
        else:  # Create new performance data frame for selector, write to file later
            self.selector_performance_data = performance_data.clone()
            self.selector_performance_data.add_solver(
                SelectionScenario.__selector_solver_name__
            )

        if isinstance(performance_data, PerformanceDataFrame):  # Convert
            # Convert the dataframes to Selector Format
            new_column_names: list[str] = []
            for solver, config_id, _ in performance_data.columns:
                if f"{solver}_{config_id}" not in new_column_names:
                    new_column_names.append(f"{solver}_{config_id}")
            self.performance_data = performance_data.drop(
                [PerformanceDataFrame.column_seed], axis=1, level=2
            )
            self.performance_data = self.performance_data.droplevel(
                [
                    PerformanceDataFrame.column_configuration,
                    PerformanceDataFrame.column_meta,
                ],
                axis=1,
            )
            self.performance_data = self.performance_data.droplevel(
                PerformanceDataFrame.index_objective, axis=0
            )
            self.performance_data.columns = new_column_names
            # Requires instances as index for both, columns as features / solvers
            # TODO: This should be an aggregation instead?
            self.performance_data.index = self.performance_data.index.droplevel("Run")
            # Enforce data type to be numeric
            self.performance_data = self.performance_data.astype(float)
            self.performance_target_path = self.directory / "performance_data.csv"
        else:  # Read from Path
            self.performance_data: pd.DataFrame = pd.read_csv(
                performance_data, index_col=0
            )
            self.performance_target_path: Path = performance_data

        if isinstance(feature_data, FeatureDataFrame):  # Convert
            self.feature_extractors = feature_data.extractors
            # Features requires instances as index, columns as feature names
            feature_target = feature_data.copy()
            feature_target.index = feature_target.index.map("_".join)  # Reduce Index
            # ASF -> feature columns, instance rows
            self.feature_data: pd.DataFrame = feature_target.T.astype(float)
            self.feature_target_path: Path = self.directory / "feature_data.csv"
        else:  # Read from Path
            self.feature_extractors = feature_extractors
            self.feature_data: pd.DataFrame = pd.read_csv(feature_data)
            self.feature_target_path: Path = feature_data

        self.ablation_scenarios: list[SelectionScenario] = []
        if ablate and len(self.performance_data.columns) > 2:
            for solver in self.performance_data.columns:
                solver_key, conf_id = solver.split("_", maxsplit=1)
                ablate_subdir = Path(f"ablated_{Path(solver).name}")
                ablated_directory = self.directory / ablate_subdir
                if (ablated_directory / "performance_data.csv").exists():
                    ablated_pd = ablated_directory / "performance_data.csv"
                elif isinstance(performance_data, PerformanceDataFrame):
                    ablated_pd = performance_data.clone()
                    ablated_pd.remove_configuration(solver_key, conf_id)
                else:  # Note we could do this but it would be hacky?
                    raise ValueError(
                        "Cannot ablate scenario after loading from file! "
                        "Requires original PerformanceDataFrame."
                    )

                self.ablation_scenarios.append(
                    SelectionScenario(
                        parent_directory=self.directory,
                        selector=selector,
                        objective=objective,
                        performance_data=ablated_pd,
                        feature_data=feature_data,
                        solver_cutoff=solver_cutoff,
                        ablate=False,  # If we set to true here, recursion would happen
                        subdir_path=ablate_subdir,
                    )
                )

    @property
    def training_instances(self: SelectionScenario) -> list[str]:
        """Get the training instances."""
        return self.performance_data.index.to_list()

    @property
    def test_instances(self: SelectionScenario) -> list[str]:
        """Get the test instances."""
        instances = self.selector_performance_data.instances
        return [i for i in instances if i not in self.training_instances]

    @property
    def training_instance_sets(self: SelectionScenario) -> list[str]:
        """Get the training instance sets."""
        # NOTE: This no longer works as instances no longer have their set in the name
        return list(set(Path(i).parent.name for i in self.training_instances))

    @property
    def test_instance_sets(self: SelectionScenario) -> list[str]:
        """Get the test instance sets."""
        # NOTE: This no longer works as instances no longer have their set in the name
        return list(set(Path(i).parent.name for i in self.test_instances))

    @property
    def instance_sets(self: SelectionScenario) -> list[str]:
        """Get all the instance sets used in this scenario."""
        return list(
            set(Path(i).parent.name for i in self.selector_performance_data.instances)
        )

    @property
    def solvers(self: SelectionScenario) -> list[str]:
        """Get the solvers used for the selector."""
        return self.performance_data.columns.to_list()

    def create_scenario(self: SelectionScenario) -> None:
        """Prepare the scenario directories."""
        self.directory.mkdir(parents=True, exist_ok=True)
        self.performance_data.to_csv(self.performance_target_path)
        self.feature_data.to_csv(self.feature_target_path)
        self.selector_performance_data.save_csv(self.selector_performance_path)
        self.create_scenario_file()

    def create_scenario_file(self: SelectionScenario) -> None:
        """Create the scenario file.

        Write the scenario to file.
        """
        with self.scenario_file.open("w") as fout:
            fout.write(self.serialise())

    def serialise(self: SelectionScenario) -> dict:
        """Serialize the scenario."""
        return (
            f"selector: {self.selector.name}\n"
            f"solver_cutoff: {self.solver_cutoff}\n"
            f"extractor_cutoff: {self.extractor_cutoff}\n"
            f"ablate: {len(self.ablation_scenarios) > 0}\n"
            f"objective: {self.objective}\n"
            f"selector_performance_data: {self.selector_performance_path}\n"
            f"performance_data: {self.performance_target_path}\n"
            f"feature_data: {self.feature_target_path}\n"
            f"feature_extractors: {','.join(self.feature_extractors)}\n"
        )

    @staticmethod
    def from_file(scenario_file: Path) -> SelectionScenario:
        """Reads scenario file and initalises SelectorScenario."""
        if not scenario_file.is_file() and (scenario_file / "scenario.txt").is_file():
            scenario_file = scenario_file / "scenario.txt"  # Resolve from directory
        values = {
            key: value.strip()
            for key, value in [
                line.split(": ", maxsplit=1) for line in scenario_file.open()
            ]
        }
        selector_class, selector_model = values["selector"].split("_", maxsplit=1)
        import ast

        selector = Selector(selector_class, selector_model)
        return SelectionScenario(
            parent_directory=scenario_file.parent,
            selector=selector,
            objective=resolve_objective(values["objective"]),
            performance_data=Path(values["performance_data"]),
            feature_data=Path(values["feature_data"]),
            feature_extractors=values["feature_extractors"].split(","),
            solver_cutoff=float(values["solver_cutoff"]),
            ablate=ast.literal_eval(values["ablate"]),
        )



"""File to help with RunRunner jobs."""

from pathlib import Path

from runrunner.base import Status
from runrunner.slurm import SlurmRun


def get_runs_from_file(
    path: Path, print_error: bool = False, filter: list[Status] = None
) -> list[SlurmRun]:
    """Retrieve all run objects from file storage.

    Args:
        path: Path object where to look recursively for the files.
        print_error: Whether to print errors.
        filter: If not None, only runs with the given statuses will be
            returned.

    Returns:
        List of all found SlumRun objects.
    """
    runs = []
    for file in path.rglob("*.json"):
        # TODO: RunRunner should be adapted to have more general methods for runs
        # So this method can work for both local and slurm
        try:
            run_obj = SlurmRun.from_file(file)
            if filter is None or run_obj.status in filter:
                runs.append(run_obj)
        except Exception as ex:
            # Not a (correct) RunRunner JSON file
            if print_error:
                print(f"[WARNING] Could not load file: {file}. Exception: {ex}")
    return runs



"""Abstract class Sparkle Callable."""

from __future__ import annotations
from pathlib import Path


class SparkleCallable:
    """Sparkle Callable class."""

    def __init__(
        self: SparkleCallable, directory: Path, runsolver_exec: Path = None
    ) -> None:
        """Initialize callable.

        Args:
            directory: Directory of the callable.
            runsolver_exec: Path to the runsolver executable.
                By default, runsolver in solver_directory.
        """
        self.directory = directory
        self.name = directory.name
        self._runsolver_exec = runsolver_exec

    @property
    def runsolver_exec(self: SparkleCallable) -> Path:
        """Return the path of the runsolver executable."""
        if self._runsolver_exec is None:
            return self.directory / "runsolver"
        return self._runsolver_exec

    def build_cmd(self: SparkleCallable) -> list[str | Path]:
        """A method that builds the commandline call string."""
        return NotImplementedError

    def run(self: SparkleCallable) -> None:
        """A method that runs the callable."""
        return NotImplementedError



"""Parameter Configuration Space tools."""

from __future__ import annotations
import re
import ast
from enum import Enum
from pathlib import Path

import ConfigSpace.conditions
import tabulate
import ConfigSpace
from ConfigSpace import ConfigurationSpace
from sparkle.tools.configspace import expression_to_configspace


class PCSConvention(Enum):
    """Internal pcs convention enum."""

    UNKNOWN = "UNKNOWN"
    SMAC = "smac"
    ParamILS = "paramils"
    IRACE = "irace"
    ConfigSpace = "configspace"


class PCSConverter:
    """Parser class independent file of notation."""

    section_regex = re.compile(r"\[(?P<name>[a-zA-Z]+?)\]\s*(?P<comment>#.*)?$")
    illegal_param_name = re.compile(r"[!:\-+@!#$%^&*()=<>?/\|~` ]")

    smac2_params_regex = re.compile(
        r"^(?P<name>[a-zA-Z0-9_]+)\s+(?P<type>[a-zA-Z]+)\s+"
        r"(?P<values>[a-zA-Z0-9\-\[\]{}_,. ]+)\s*"
        r"\[(?P<default>[a-zA-Z0-9._-]+)\]?\s*"
        r"(?P<scale>log)?\s*(?P<comment>#.*)?$"
    )
    smac2_conditions_regex = re.compile(
        r"^(?P<parameter>[a-zA-Z0-9_]+)\s*\|\s*"
        r"(?P<expression>.+)$"
    )
    smac2_forbidden_regex = re.compile(r"\{(?P<forbidden>.+)\}$")

    paramils_params_regex = re.compile(
        r"^(?P<name>[a-zA-Z0-9@!#:_-]+)\s*"
        r"(?P<values>{[a-zA-Z0-9._+\-\, ]+})\s*"
        r"\[(?P<default>[a-zA-Z0-9._\-+ ]+)\]?\s*"
        r"(?P<comment>#.*)?$"
    )
    paramils_conditions_regex = re.compile(
        r"^(?P<parameter>[a-zA-Z0-9@!#:_-]+)\s*\|\s*"
        r"(?P<expression>.+)$"
    )

    irace_params_regex = re.compile(
        r"^(?P<name>[a-zA-Z0-9_]+)\s+"
        r"(?P<switch>[\"a-zA-Z0-9_\- ]+)\s+"
        r"(?P<type>[cior])(?:,)?(?P<scale>log)?\s+"
        r"(?P<values>[a-zA-Z0-9\-()_,. ]+)\s*"
        r"(?:\|)?(?P<conditions>[a-zA-Z-0-9_!=<>\%()\&\|\. ]*)?\s*"
        r"(?P<comment>#.*)?$"
    )

    @staticmethod
    def get_convention(file: Path) -> PCSConvention:
        """Determines the format of a pcs file."""
        try:
            ConfigSpace.ConfigurationSpace.from_yaml(file)
            return PCSConvention.ConfigSpace
        except Exception:
            pass
        try:
            ConfigSpace.ConfigurationSpace.from_json(file)
            return PCSConvention.ConfigSpace
        except Exception:
            pass

        file_contents = file.open().readlines()
        for line in file_contents:
            if line.startswith("#"):  # Comment line
                continue
            if "#" in line:
                line, _ = line.split("#", maxsplit=1)
            line = line.strip()
            if re.match(PCSConverter.smac2_params_regex, line):
                return PCSConvention.SMAC
            elif re.match(PCSConverter.paramils_params_regex, line):
                return PCSConvention.ParamILS
            elif re.match(PCSConverter.irace_params_regex, line):
                return PCSConvention.IRACE
        return PCSConvention.UNKNOWN

    @staticmethod
    def parse(file: Path, convention: PCSConvention = None) -> ConfigurationSpace:
        """Determines the format of a pcs file and parses into Configuration Space."""
        if not convention:
            convention = PCSConverter.get_convention(file)
        if convention == PCSConvention.ConfigSpace:
            if file.suffix == ".yaml":
                return ConfigSpace.ConfigurationSpace.from_yaml(file)
            if file.suffix == ".json":
                return ConfigSpace.ConfigurationSpace.from_json(file)
        if convention == PCSConvention.SMAC:
            return PCSConverter.parse_smac(file)
        if convention == PCSConvention.ParamILS:
            return PCSConverter.parse_paramils(file)
        if convention == PCSConvention.IRACE:
            return PCSConverter.parse_irace(file)
        raise Exception(
            f"PCS convention not recognised based on any lines in file:\n{file}"
        )

    @staticmethod
    def parse_smac(content: list[str] | Path) -> ConfigurationSpace:
        """Parses a SMAC2 file."""
        space_name = content.name if isinstance(content, Path) else None
        content = content.open().readlines() if isinstance(content, Path) else content
        cs = ConfigurationSpace(space_name)
        for line in content:
            if not line.strip() or line.startswith("#"):  # Empty or comment
                continue
            comment = None
            line = line.strip()
            if re.match(PCSConverter.smac2_params_regex, line):
                parameter = re.fullmatch(PCSConverter.smac2_params_regex, line)
                name = parameter.group("name")
                parameter_type = parameter.group("type")
                values = parameter.group("values")
                default = parameter.group("default")
                comment = parameter.group("comment")
                scale = parameter.group("scale")
                if parameter_type == "integer":
                    values = ast.literal_eval(values)
                    csparam = ConfigSpace.UniformIntegerHyperparameter(
                        name=name,
                        lower=int(values[0]),
                        upper=int(values[-1]),
                        default_value=int(default),
                        log=scale == "log",
                        meta=comment,
                    )
                elif parameter_type == "real":
                    values = ast.literal_eval(values)
                    csparam = ConfigSpace.UniformFloatHyperparameter(
                        name=name,
                        lower=float(values[0]),
                        upper=float(values[-1]),
                        default_value=float(default),
                        log=scale == "log",
                        meta=comment,
                    )
                elif parameter_type == "categorical":
                    values = re.sub(r"[{}\s]+", "", values).split(",")
                    csparam = ConfigSpace.CategoricalHyperparameter(
                        name=name,
                        choices=values,
                        default_value=default,
                        meta=comment,
                        # Does not seem to contain any weights?
                    )
                elif parameter_type == "ordinal":
                    values = re.sub(r"[{}\s]+", "", values).split(",")
                    csparam = ConfigSpace.OrdinalHyperparameter(
                        name=name,
                        sequence=values,
                        default_value=default,
                        meta=comment,
                    )
                cs.add(csparam)
            elif re.match(PCSConverter.smac2_conditions_regex, line):
                # Break up the expression into the smallest possible pieces
                match = re.fullmatch(PCSConverter.smac2_conditions_regex, line)
                parameter, condition = (
                    match.group("parameter"),
                    match.group("expression"),
                )
                parameter = cs[parameter.strip()]
                condition = condition.replace(" || ", " or ").replace(" && ", " and ")
                condition = re.sub(r"(?<![<>!=])=(?<![=])", "==", condition)
                condition = re.sub(r"!==", "!=", condition)
                condition = expression_to_configspace(
                    condition, cs, target_parameter=parameter
                )
                cs.add(condition)
            elif re.match(PCSConverter.smac2_forbidden_regex, line):
                match = re.fullmatch(PCSConverter.smac2_forbidden_regex, line)
                forbidden = match.group("forbidden")
                # Forbidden expressions structure <expression> <operator> <value>
                # where expressions can contain:
                # Logical Operators: >=, <=, >, <, ==, !=,
                # Logical clause operators: ( ), ||, &&,
                # Supported by SMAC2 but not by ConfigSpace?:
                # Arithmetic Operators: +, -, *, ^, %
                # Functions: abs, acos, asin, atan, cbrt, ceil, cos, cosh, exp, floor,
                # log, log10, log2, sin, sinh, sqrt, tan, tanh
                # NOTE: According to MA & JR, these were never actually supported
                rejected_operators = (
                    "+",
                    "-",
                    "*",
                    "^",
                    "%",
                    "abs",
                    "acos",
                    "asin",
                    "atan",
                    "cbrt",
                    "ceil",
                    "cos",
                    "cosh",
                    "exp",
                    "floor",
                    "log",
                    "log10",
                    "log2",
                    "sin",
                    "sinh",
                    "sqrt",
                    "tan",
                    "tanh",
                )
                if any([r in forbidden.split(" ") for r in rejected_operators]):
                    print(
                        "WARNING: Arithmetic operators are not supported by "
                        "ConfigurationSpace. Skipping forbidden expression:\n"
                        f"{forbidden}"
                    )
                    continue
                forbidden = (
                    forbidden.replace(" && ", " and ")
                    .replace(", ", " and ")
                    .replace(" || ", " or ")
                    .strip()
                )  # To AST notation
                forbidden = re.sub(r"(?<![<>!=])=(?![=])", "==", forbidden)
                forbidden = expression_to_configspace(forbidden, cs)
                cs.add(forbidden)
            else:
                raise Exception(f"SMAC2 PCS expression not recognised on line:\n{line}")
        return cs

    @staticmethod
    def parse_paramils(content: list[str] | Path) -> ConfigurationSpace:
        """Parses a paramils file."""
        space_name = content.name if isinstance(content, Path) else None
        content = content.open().readlines() if isinstance(content, Path) else content
        cs = ConfigurationSpace(name=space_name)
        conditions_lines = {}
        for line in content:
            line = line.strip()
            if not line or line.startswith("#"):  # Empty or comment
                continue
            comment = None
            if re.match(PCSConverter.paramils_params_regex, line):
                parameter = re.fullmatch(PCSConverter.paramils_params_regex, line)
                name = parameter.group("name")
                if re.match(PCSConverter.illegal_param_name, name):
                    # ParamILS is flexible to which parameters are allowed.
                    # We do not allow it as it creates many problems with parsing
                    # expressions
                    raise ValueError(
                        f"ParamILS parameter name not allowed: {name}. "
                        "This is supported by ParamILS, but not by PCSConverter."
                    )
                values = parameter.group("values")
                values = values.replace("..", ",")  # Replace automatic expansion
                try:
                    values = list(ast.literal_eval(values))  # Values are sets
                    values = sorted(values)
                    if any([isinstance(v, float) for v in values]):
                        parameter_type = float
                    else:
                        parameter_type = int
                except Exception:  # of strings (Categorical)
                    values = values.replace("{", "").replace("}", "").split(",")
                    parameter_type = str
                if len(values) == 1:  # Not allowed by ConfigSpace for int / float
                    values = [str(values[0])]
                    parameter_type = str
                default = parameter.group("default")
                comment = parameter.group("comment")
                if parameter_type is int:
                    csparam = ConfigSpace.UniformIntegerHyperparameter(
                        name=name,
                        lower=int(values[0]),
                        upper=int(values[-1]),
                        default_value=int(default),
                        meta=comment,
                    )
                elif parameter_type is float:
                    csparam = ConfigSpace.UniformFloatHyperparameter(
                        name=name,
                        lower=float(values[0]),
                        upper=float(values[-1]),
                        default_value=float(default),
                        meta=comment,
                    )
                elif parameter_type is str:
                    csparam = ConfigSpace.CategoricalHyperparameter(
                        name=name,
                        choices=values,
                        default_value=default,
                        meta=comment,
                    )
                cs.add(csparam)
            elif re.match(PCSConverter.paramils_conditions_regex, line):
                # Break up the expression into the smallest possible pieces
                match = re.fullmatch(PCSConverter.paramils_conditions_regex, line)
                parameter, condition = (
                    match.group("parameter").strip(),
                    match.group("expression"),
                )
                condition = condition.replace(" || ", " or ").replace(" && ", " and ")
                condition = re.sub(r"(?<![<>!=])=(?<![=])", "==", condition)
                condition = re.sub(r"!==", "!=", condition)
                # ParamILS supports multiple lines of conditions for a single parameter
                # so we collect, with the AND operator and parse + add them later
                if parameter not in conditions_lines:
                    conditions_lines[parameter] = condition
                else:
                    conditions_lines[parameter] += " and " + condition
            elif re.match(PCSConverter.smac2_forbidden_regex, line):
                match = re.fullmatch(PCSConverter.smac2_forbidden_regex, line)
                forbidden = match.group("forbidden")
                # Forbidden expressions structure <expression> <operator> <value>
                # where expressions can contain:
                # Logical Operators: >=, <=, >, <, ==, !=,
                # Logical clause operators: ( ), ||, &&,
                # Supported by SMAC2 but not by ConfigSpace?:
                # Arithmetic Operators: +, -, *, ^, %
                # Functions: abs, acos, asin, atan, cbrt, ceil, cos, cosh, exp, floor,
                # log, log10, log2, sin, sinh, sqrt, tan, tanh
                # NOTE: According to MA & JR, these were never actually supported
                rejected_operators = (
                    "+",
                    "-",
                    "*",
                    "^",
                    "%",
                    "abs",
                    "acos",
                    "asin",
                    "atan",
                    "cbrt",
                    "ceil",
                    "cos",
                    "cosh",
                    "exp",
                    "floor",
                    "log",
                    "log10",
                    "log2",
                    "sin",
                    "sinh",
                    "sqrt",
                    "tan",
                    "tanh",
                )
                if any([r in forbidden.split(" ") for r in rejected_operators]):
                    print(
                        "WARNING: Arithmetic operators are not supported by "
                        "ConfigurationSpace. Skipping forbidden expression:\n"
                        f"{forbidden}"
                    )
                    continue
                forbidden = (
                    forbidden.replace(" && ", " and ")
                    .replace(", ", " and ")
                    .replace(" || ", " or ")
                    .strip()
                )  # To AST notation
                forbidden = re.sub(r"(?<![<>!=])=(?![=])", "==", forbidden)
                forbidden = expression_to_configspace(forbidden, cs)
                cs.add(forbidden)
            else:
                raise Exception(
                    f"ParamILS PCS expression not recognised on line: {line}"
                )
        # Add the condition
        for pname, cond in conditions_lines.items():  # Add conditions
            condition = expression_to_configspace(cond, cs, target_parameter=cs[pname])
            cs.add(condition)
        return cs

    @staticmethod
    def parse_irace(content: list[str] | Path) -> ConfigurationSpace:
        """Parses a irace file."""
        space_name = content.name if isinstance(content, Path) else None
        content = content.open().readlines() if isinstance(content, Path) else content
        cs = ConfigurationSpace(name=space_name)
        standardised_conditions = []
        forbidden_flag, global_flag = False, False
        for line in content:
            line = line.strip()
            if not line or line.startswith("#"):  # Empty or comment
                continue
            if re.match(PCSConverter.section_regex, line):
                section = re.fullmatch(PCSConverter.section_regex, line)
                if section.group("name") == "forbidden":
                    forbidden_flag, global_flag = True, False
                    continue
                elif section.group("name") == "global":
                    global_flag, forbidden_flag = True, False
                    continue
                else:
                    raise Exception(f"IRACE PCS section not recognised on line:\n{line}")
            elif global_flag:  # Parse global statements
                continue  # We do not parse global group
            elif forbidden_flag:  # Parse forbidden statements
                # Parse the forbidden statement to standardised format
                forbidden_expr = re.sub(r" \& ", " and ", line)
                forbidden_expr = re.sub(r" \| ", " or ", forbidden_expr)
                forbidden_expr = re.sub(r" \%in\% ", " in ", forbidden_expr)
                forbidden_expr = re.sub(r" [co]\(", " (", forbidden_expr)
                forbidden_expr = expression_to_configspace(forbidden_expr, cs)
                cs.add(forbidden_expr)
            elif re.match(PCSConverter.irace_params_regex, line):
                parameter = re.fullmatch(PCSConverter.irace_params_regex, line)
                name = parameter.group("name")
                parameter_type = parameter.group("type")
                # NOTE: IRACE supports depedent parameter domains, e.g. parameters which
                # domain relies on another parameter: p2 "--p2" r ("p1", "p1 + 10")"
                # and is limited to the operators: +,-, *, /, %%, min, max
                values = ast.literal_eval(parameter.group("values"))
                scale = parameter.group("scale")
                conditions = parameter.group("conditions")
                comment = parameter.group("comment")
                # Convert categorical / ordinal values to strings
                if parameter_type == "c" or parameter_type == "o":
                    values = [str(i) for i in values]
                else:
                    if any(
                        operator in values
                        for operator in ["+", "-", "*", "/", "%", "min", "max"]
                    ):
                        raise ValueError(
                            "Dependent parameter domains not supported by "
                            "ConfigurationSpace."
                        )
                    lower_bound, upper_bound = values[0], values[1]
                if parameter_type == "c":
                    csparam = ConfigSpace.CategoricalHyperparameter(
                        name=name,
                        choices=values,
                        meta=comment,
                    )
                elif parameter_type == "o":
                    csparam = ConfigSpace.OrdinalHyperparameter(
                        name=name,
                        sequence=values,
                        meta=comment,
                    )
                elif parameter_type == "r":
                    csparam = ConfigSpace.UniformFloatHyperparameter(
                        name=name,
                        lower=float(lower_bound),
                        upper=float(upper_bound),
                        log=scale == "log",
                        meta=comment,
                    )
                elif parameter_type == "i":
                    csparam = ConfigSpace.UniformIntegerHyperparameter(
                        name=name,
                        lower=int(lower_bound),
                        upper=int(upper_bound),
                        log=scale == "log",
                        meta=comment,
                    )
                cs.add(csparam)
                if conditions:
                    # Convert the expression to standardised format
                    conditions = re.sub(r" \& ", " and ", conditions)
                    conditions = re.sub(r" \| ", " or ", conditions)
                    conditions = re.sub(r" \%in\% ", " in ", conditions)
                    conditions = re.sub(r" [cior]\(", " (", conditions)
                    conditions = conditions.strip()
                    standardised_conditions.append((csparam, conditions))
            else:
                raise Exception(f"IRACE PCS expression not recognised on line:\n{line}")

        # We can only add the conditions after all parameters have been parsed:
        for csparam, conditions in standardised_conditions:
            conditions = expression_to_configspace(
                conditions, cs, target_parameter=csparam
            )
            cs.add(conditions)
        return cs

    @staticmethod
    def export(
        configspace: ConfigurationSpace, pcs_format: PCSConvention, file: Path
    ) -> str | None:
        """Exports a config space object to a specific PCS convention.

        Args:
            configspace: ConfigurationSpace, the space to convert
            pcs_format: PCSConvention, the convention to conver to
            file: Path, the file to write to. If None, will return string.

        Returns:
            String in case of no file path given, otherwise None.
        """
        # Create pcs table
        declaration = (
            f"### {pcs_format.name} Parameter Configuration Space file "
            "generated by Sparkle\n"
        )
        rows = []
        extra_rows = []
        if pcs_format == PCSConvention.SMAC or pcs_format == PCSConvention.ParamILS:
            import numpy as np

            granularity = 20  # For ParamILS. TODO: Make it parametrisable
            header = [
                "# Parameter Name",
                "type",
                "values",
                "default value",
                "scale",
                "comments",
            ]
            parameter_map = {
                ConfigSpace.UniformFloatHyperparameter: "real",
                ConfigSpace.UniformIntegerHyperparameter: "integer",
                ConfigSpace.CategoricalHyperparameter: "categorical",
                ConfigSpace.OrdinalHyperparameter: "ordinal",
            }
            for parameter in list(configspace.values()):
                log = False
                if isinstance(
                    parameter, ConfigSpace.hyperparameters.NumericalHyperparameter
                ):
                    log = parameter.log
                    if pcs_format == PCSConvention.ParamILS:  # Discretise
                        dtype = (
                            float
                            if isinstance(
                                parameter, ConfigSpace.UniformFloatHyperparameter
                            )
                            else int
                        )
                        if log:
                            lower = 1e-5 if parameter.lower == 0 else parameter.lower
                            domain = list(
                                np.unique(
                                    np.geomspace(
                                        lower,
                                        parameter.upper,
                                        granularity,
                                        dtype=dtype,
                                    )
                                )
                            )
                        else:
                            domain = list(
                                np.linspace(
                                    parameter.lower,
                                    parameter.upper,
                                    granularity,
                                    dtype=dtype,
                                )
                            )
                        if dtype(parameter.default_value) not in domain:  # Add default
                            domain += [dtype(parameter.default_value)]
                        domain = list(set(domain))  # Ensure unique values only
                        domain.sort()
                        domain = "{" + ",".join([str(i) for i in domain]) + "}"
                    else:  # SMAC2 takes ranges
                        domain = f"[{parameter.lower}, {parameter.upper}]"
                else:
                    domain = "{" + ",".join(parameter.choices) + "}"
                rows.append(
                    [
                        parameter.name,
                        parameter_map[type(parameter)]
                        if not pcs_format == PCSConvention.ParamILS
                        else "",
                        domain,
                        f"[{parameter.default_value}]",
                        "log"
                        if log and not pcs_format == PCSConvention.ParamILS
                        else "",
                        f"# {parameter.meta}",
                    ]
                )
            if configspace.conditions:
                extra_rows.extend(["", "# Parameter Conditions"])
                for condition in configspace.conditions:
                    condition_str = str(condition)
                    condition_str = condition_str.replace("(", "").replace(
                        ")", ""
                    )  # Brackets not allowed
                    condition_str = condition_str.replace("'", "")  # No quotes needed
                    condition_str = condition_str.replace(
                        f"{condition.child.name} | ", ""
                    ).strip()
                    condition_str = f"{condition.child.name} | " + condition_str
                    if pcs_format == PCSConvention.ParamILS and re.search(
                        r"[<>!=]+|[<>]=|[!=]=", condition_str
                    ):
                        # TODO: Translate condition ParamILS expression (in)
                        continue
                    extra_rows.append(condition_str)
            if configspace.forbidden_clauses:
                extra_rows.extend(["", "# Forbidden Expressions"])
                for forbidden in configspace.forbidden_clauses:
                    forbidden_str = str(forbidden).replace("Forbidden: ", "")
                    forbidden_str = forbidden_str.replace("(", "{").replace(")", "}")
                    forbidden_str = forbidden_str.replace("'", "")
                    if pcs_format == PCSConvention.ParamILS and re.search(
                        r"[<>!=]+|[<>]=|[!=]=", forbidden_str
                    ):
                        # TODO: Translate condition ParamILS expression (in)
                        continue
                    extra_rows.append(forbidden_str)
        elif pcs_format == PCSConvention.IRACE:
            digits = 4  # Number of digits after decimal point required
            parameter_map = {
                ConfigSpace.UniformFloatHyperparameter: "r",
                ConfigSpace.UniformIntegerHyperparameter: "i",
                ConfigSpace.CategoricalHyperparameter: "c",
                ConfigSpace.OrdinalHyperparameter: "o",
            }
            header = [
                "# Parameter Name",
                "switch",
                "type",
                "values",
                "[conditions (using R syntax)]",
                "comments",
            ]
            for parameter in list(configspace.values()):
                parameter_conditions = []
                for c in configspace.conditions:
                    if c.child == parameter:
                        parameter_conditions.append(c)
                parameter_type = parameter_map[type(parameter)]
                condition_type = (
                    parameter_type
                    if type(parameter) is ConfigSpace.CategoricalHyperparameter
                    else ""
                )
                condition_str = " || ".join([str(c) for c in parameter_conditions])
                condition_str = condition_str.replace(f"{parameter.name} | ", "")
                condition_str = condition_str.replace(" in ", f" %in% {condition_type}")
                condition_str = condition_str.replace("{", "(").replace("}", ")")
                condition_str = condition_str.replace("'", "")  # No quotes around string
                condition_str = condition_str.replace(" && ", " & ").replace(
                    " || ", " | "
                )
                if isinstance(
                    parameter, ConfigSpace.hyperparameters.NumericalHyperparameter
                ):
                    if parameter.log:
                        parameter_type += ",log"
                    domain = f"({parameter.lower}, {parameter.upper})"
                    if isinstance(
                        parameter, ConfigSpace.hyperparameters.FloatHyperparameter
                    ):
                        # Format the floats to interpret the number of digits
                        # (Includes scientific notation)
                        lower, upper = (
                            format(parameter.lower, ".16f").strip("0"),
                            format(parameter.upper, ".16f").strip("0"),
                        )
                        param_digits = max(
                            len(str(lower).split(".")[1]),
                            len(str(upper).split(".")[1]),
                        )
                        # Check if we need to update the global digits
                        if param_digits > digits:
                            digits = param_digits
                else:
                    domain = "(" + ",".join(parameter.choices) + ")"
                rows.append(
                    [
                        parameter.name,
                        f'"--{parameter.name} "',
                        parameter_type,
                        domain,  # Parameter range/domain
                        f"| {condition_str}" if condition_str else "",
                        f"# {parameter.meta}" if parameter.meta else "",
                    ]
                )
            if configspace.forbidden_clauses:
                extra_rows.extend(["", "[forbidden]"])
                for forbidden_expression in configspace.forbidden_clauses:
                    forbidden_str = str(forbidden_expression).replace("Forbidden: ", "")
                    if " in " in forbidden_str:
                        type_char = parameter_map[
                            type(forbidden_expression.hyperparameter)
                        ]
                        forbidden_str.replace(" in ", f" %in% {type_char}")
                    forbidden_str = forbidden_str.replace(" && ", " & ").replace(
                        " || ", " | "
                    )
                    extra_rows.append(forbidden_str)
            if digits > 4:  # Default digits is 4
                extra_rows.extend(["", "[global]", f"digits={digits}"])

        output = (
            declaration
            + tabulate.tabulate(rows, headers=header, tablefmt="plain", numalign="left")
            + "\n"
        )
        if extra_rows:
            output += "\n".join(extra_rows) + "\n"
        if file is None:
            return output
        file.open("w+").write(output)

    @staticmethod
    def validate(file_path: Path) -> bool:
        """Validate a pcs file."""
        # TODO: Determine which format
        # TODO: Verify each line, and the order in which they were written
        return



#!/usr/bin/env python3
"""Configurator CLI wrapper class to standardise I/O of Configurators."""

import sys
import subprocess
from pathlib import Path
import implementations

if __name__ == "__main__":
    args = sys.argv
    configurator_name = args[1]
    output_source = Path(args[2])
    output_target = Path(args[3])
    scenario = Path(args[4])
    configuration_id = args[5]
    configurator_call = args[6:]
    # 1. Resolve for which Configurator we are standardising
    configurator = implementations.resolve_configurator(configurator_name)
    scenario = configurator.scenario_class().from_file(scenario)
    # 2. Execute the call, output is automatically piped to the caller's set output
    subprocess.run(configurator_call)
    # 3. Have the configurator implementation organise the output, standardised
    configurator.organise_output(
        output_source=output_source,
        output_target=output_target,
        scenario=scenario,
        configuration_id=configuration_id,
    )
    print(f"Organising done! See {output_target} for results.")



"""Module to manage feature data files and common operations on them."""

from __future__ import annotations
import pandas as pd
import math
from pathlib import Path


class FeatureDataFrame(pd.DataFrame):
    """Class to manage feature data CSV files and common operations on them."""

    missing_value = math.nan
    multi_dim_names = ["FeatureGroup", "FeatureName", "Extractor"]

    def __init__(
        self: FeatureDataFrame,
        csv_filepath: Path,
        instances: list[str] = [],
        extractor_data: dict[str, list[tuple[str, str]]] = {},
    ) -> None:
        """Initialise a FeatureDataFrame object.

        Arguments:
            csv_filepath: The Path for the CSV storage. If it does not exist,
                a new DataFrame will be initialised and stored here.
            instances: The list of instances (Columns) to be added to the DataFrame.
            extractor_data: A dictionary with extractor names as key, and a list of
                tuples ordered as [(feature_group, feature_name), ...] as value.
        """
        # Initialize a dataframe from an existing file
        if csv_filepath.exists():
            # Read from file
            temp_df = pd.read_csv(
                csv_filepath, index_col=FeatureDataFrame.multi_dim_names
            )
            super().__init__(temp_df)
            self.csv_filepath = csv_filepath
        # Create a new dataframe
        else:
            # Unfold the extractor_data into lists
            multi_index_lists = [[], [], []]
            for extractor in extractor_data:
                for group, feature_name in extractor_data[extractor]:
                    multi_index_lists[0].append(group)
                    multi_index_lists[1].append(feature_name)
                    multi_index_lists[2].append(extractor)
            # Initialise new dataframe
            multi_index = pd.MultiIndex.from_arrays(
                multi_index_lists, names=self.multi_dim_names
            )
            super().__init__(
                data=self.missing_value, index=multi_index, columns=instances
            )
            self.csv_filepath = csv_filepath
            self.save_csv()

    def add_extractor(
        self: FeatureDataFrame,
        extractor: str,
        extractor_features: list[tuple[str, str]],
        values: list[list[float]] = None,
    ) -> None:
        """Add an extractor and its feature names to the dataframe.

        Arguments:
            extractor: Name of the extractor
            extractor_features: Tuples of [FeatureGroup, FeatureName]
            values: Initial values of the Extractor per instance in the dataframe.
                Defaults to FeatureDataFrame.missing_value.
        """
        if values is None:
            values = [self.missing_value] * len(extractor_features)
        # Unfold to indices to lists
        for index, pair in enumerate(extractor_features):
            feature_group, feature = pair
            self.loc[(feature_group, feature, extractor), :] = values[index]

    def add_instances(
        self: FeatureDataFrame, instance: str | list[str], values: list[float] = None
    ) -> None:
        """Add one or more instances to the dataframe."""
        if values is None:
            values = FeatureDataFrame.missing_value
        self[instance] = values

    def remove_extractor(self: FeatureDataFrame, extractor: str) -> None:
        """Remove an extractor from the dataframe."""
        self.drop(extractor, axis=0, level="Extractor", inplace=True)

    def remove_instances(self: FeatureDataFrame, instances: str | list[str]) -> None:
        """Remove an instance from the dataframe."""
        self.drop(instances, axis=1, inplace=True)

    def get_feature_groups(
        self: FeatureDataFrame, extractor: str | list[str] = None
    ) -> list[str]:
        """Retrieve the feature groups in the dataframe.

        Args:
            extractor: Optional. If extractor(s) are given,
                yields only feature groups of that extractor.

        Returns:
            A list of feature groups.
        """
        indices = self.index
        if extractor is not None:
            if isinstance(extractor, str):
                extractor = [extractor]
            indices = indices[indices.isin(extractor, level=2)]
        return indices.get_level_values(level=0).unique().to_list()

    def get_value(
        self: FeatureDataFrame,
        instance: str,
        extractor: str,
        feature_group: str,
        feature_name: str,
    ) -> None:
        """Return a value in the dataframe."""
        return self.loc[(feature_group, feature_name, extractor), instance]

    def set_value(
        self: FeatureDataFrame,
        instance: str,
        extractor: str,
        feature_group: str,
        feature_name: str,
        value: float,
    ) -> None:
        """Set a value in the dataframe."""
        self.loc[(feature_group, feature_name, extractor), instance] = value

    def has_missing_vectors(self: FeatureDataFrame) -> bool:
        """Returns True if there are any Extractors still to be run on any instance."""
        for instance in self.columns:
            for extractor in self.extractors:
                extractor_features = self.xs(extractor, level=2, drop_level=False)
                if extractor_features.loc[:, instance].isnull().all():
                    return True
        return False

    def remaining_jobs(self: FeatureDataFrame) -> list[tuple[str, str, str]]:
        """Determines needed feature computations per instance/extractor/group.

        Returns:
            list: A list of tuples representing (Extractor, Instance, Feature Group).
                that needs to be computed.
        """
        remaining_jobs = []
        for extractor in self.extractors:
            for group in self.get_feature_groups(extractor):
                subset = self.xs((group, extractor), level=(0, 2))
                for instance in self.columns:
                    if subset.loc[:, instance].isnull().all():
                        remaining_jobs.append((instance, extractor, group))
        return remaining_jobs

    def get_instance(self: FeatureDataFrame, instance: str) -> list[float]:
        """Return the feature vector of an instance."""
        return self[instance].tolist()

    def impute_missing_values(self: FeatureDataFrame) -> None:
        """Imputes all NaN values by taking the average feature value."""
        imputed_df = self.T.fillna(self.mean(axis=1)).T
        self[:] = imputed_df.values

    def has_missing_value(self: FeatureDataFrame) -> bool:
        """Return whether there are missing values in the feature data."""
        return self.isnull().any().any()

    def reset_dataframe(self: FeatureDataFrame) -> bool:
        """Resets all values to FeatureDataFrame.missing_value."""
        self.loc[:, :] = FeatureDataFrame.missing_value

    def sort(self: FeatureDataFrame) -> None:
        """Sorts the DataFrame by Multi-Index for readability."""
        self.sort_index(level=FeatureDataFrame.multi_dim_names, inplace=True)

    @property
    def instances(self: FeatureDataFrame) -> list[str]:
        """Return the instances in the dataframe."""
        return self.columns

    @property
    def extractors(self: FeatureDataFrame) -> list[str]:
        """Returns all unique extractors in the DataFrame."""
        return self.index.get_level_values("Extractor").unique().to_list()

    @property
    def num_features(self: FeatureDataFrame) -> int:
        """Return the number of features in the dataframe."""
        return self.shape[0]

    @property
    def features(self: FeatureDataFrame) -> list[str]:
        """Return the features in the dataframe."""
        return self.index.get_level_values("FeatureName").unique().to_list()

    def save_csv(self: FeatureDataFrame, csv_filepath: Path = None) -> None:
        """Write a CSV to the given path.

        Args:
            csv_filepath: String path to the csv file. Defaults to self.csv_filepath.
        """
        csv_filepath = self.csv_filepath if csv_filepath is None else csv_filepath
        if csv_filepath is None:
            raise ValueError("Cannot save DataFrame: no `csv_filepath` was provided.")
        self.to_csv(csv_filepath)



#!/usr/bin/env python3
"""Command to interact with async jobs."""

import sys
import time
import argparse

import pytermgui as ptg
from tabulate import tabulate

from runrunner.base import Status, Run
from runrunner.slurm import SlurmRun

from sparkle.platform.cli_types import TEXT
from sparkle.CLI.help import logging
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help import jobs as jobs_help
from sparkle.CLI.help import global_variables as gv


def parser_function() -> argparse.ArgumentParser:
    """Create parser for the jobs command."""
    parser = argparse.ArgumentParser(
        description="Command to interact with async jobs. "
        "The command starts an interactive "
        "table when no flags are given. Jobs "
        "can be selected for cancelling in the "
        "table and non activate jobs can be "
        "flushed by pressing the spacebar."
    )
    parser.add_argument(*ac.CancelJobsArgument.names, **ac.CancelJobsArgument.kwargs)
    parser.add_argument(*ac.JobIDsArgument.names, **ac.JobIDsArgument.kwargs)
    parser.add_argument(*ac.AllJobsArgument.names, **ac.AllJobsArgument.kwargs)
    return parser


def create_jobs_table(
    jobs: list[SlurmRun], markup: bool = True, format: str = "grid"
) -> str:
    """Create a table of jobs.

    Args:
        jobs: List of SlurmRun objects.
        markup: By default some mark up will be applied to the table.
            If false, a more plain version will be created.
        format: The tabulate format to use.

    Returns:
        A table of jobs as a string.
    """
    job_table = [
        [
            "RunId",
            "Name",
            "Quality of Service",
            "Partition",
            "Status",
            "Dependencies",
            "Finished Jobs",
            "Run Time",
        ]
    ]
    for job in jobs:
        # Count number of jobs that have finished
        finished_jobs_count = sum(
            1 for status in job.all_status if status == Status.COMPLETED
        )
        if markup:  # Format job.status
            status_text = (
                TEXT.format_text([TEXT.BOLD], job.status)
                if job.status == Status.RUNNING
                else (
                    TEXT.format_text([TEXT.ITALIC], job.status)
                    if job.status == Status.COMPLETED
                    else job.status.value
                )
            )
        else:
            status_text = job.status.value
        job_table.append(
            [
                job.run_id,
                job.name,
                job.qos,
                job.partition,
                status_text,
                "None" if len(job.dependencies) == 0 else ", ".join(job.dependencies),
                f"{finished_jobs_count}/{len(job.all_status)}",
                job.runtime,
            ]
        )
    if markup:
        job_table = tabulate(
            job_table,
            headers="firstrow",
            tablefmt=format,
            maxcolwidths=[12, 32, 14, 12, 16, 16, 16, 10],
        )
    return job_table


def table_gui(jobs: list[Run]) -> None:
    """Display a table of running jobs."""
    jobs = sorted(jobs, key=lambda j: j.run_id)
    ptg.Button.chars = {"delimiter": ["", ""]}  # Disable padding around buttons

    def cancel_jobs(self: ptg.Button) -> None:
        """Cancel jobs based on a button click."""
        job_id = self.label.split("|")[1].strip()
        job = job_id_map[job_id]

        def kill_exit(self: ptg.Button) -> None:
            """Two step protocol of killing the job and removing the popup."""
            job.kill()
            manager.remove(popup)

        button_yes = ptg.Button("Yes", kill_exit)
        button_no = ptg.Button("No", lambda *_: manager.remove(popup))

        popup = manager.alert(ptg.Label(f"Cancel job {job_id}?"), button_no, button_yes)

        refresh_data(self.parent)

    def refresh_data(self: ptg.Window | ptg.WindowManager, key: str = None) -> None:
        """Refresh the table."""
        # Resolve window
        window = self._windows[-1] if isinstance(self, ptg.WindowManager) else self
        # Fetch latest data
        for job in jobs:
            if job.status in [Status.WAITING, Status.RUNNING]:
                job.get_latest_job_details()
        job_table = create_jobs_table(jobs, markup=True).splitlines()

        if window.width != len(job_table[0]):  # Resize window
            window.width = len(job_table[0])

        for index, row in enumerate(job_table):
            if row.startswith("|"):
                row_id = row.split("|")[1].strip()
                if row_id in job_id_map.keys() and job_id_map[row_id].status in [
                    Status.WAITING,
                    Status.RUNNING,
                ]:
                    window._widgets[index + 1] = ptg.Button(row, cancel_jobs)
                else:
                    window._widgets[index + 1] = ptg.Label(row)
            else:
                window._widgets[index + 1] = ptg.Label(row)
            window._widgets[index + 1].parent = window

    table = create_jobs_table(jobs, markup=True).splitlines()
    with ptg.WindowManager() as manager:

        def macro_reload(fmt: str) -> str:
            """Updates jobs in the table with an interval."""
            if "last_reload" not in globals():
                global last_reload
                last_reload = time.time()
            diff = time.time() - last_reload
            interval = 10.0
            if diff > interval:  # Check every 10 seconds
                last_reload = 0
                any_running = False
                for job in jobs:
                    if job.status in [Status.RUNNING, Status.WAITING]:
                        any_running = True
                        job.get_latest_job_details()
                if not any_running:
                    manager.stop()
                refresh_data(manager)
                last_reload = time.time()
            n_bars = int(diff / 2)
            return "|" + "█" * n_bars + " " * (4 - n_bars) + "|"

        def flush_popup(self: ptg.WindowManager, key: str) -> None:
            """Pop up for flushing completed jobs."""
            if len(jobs) <= 1:  # Cannot flush the last job
                return
            flushable_jobs = []
            flushable_job_ids = []
            for job in jobs:
                if job.status not in [Status.WAITING, Status.RUNNING]:
                    flushable_jobs.append(job)
                    flushable_job_ids.append(job.run_id)
            if len(flushable_jobs) == 0:  # Nothing to flush
                return

            def flush(self: ptg.Button) -> None:
                """Flush completed jobs."""
                for job in flushable_jobs:
                    jobs.remove(job)
                    del job_id_map[job.run_id]
                flushable_widgets = []
                table_window = manager._windows[-1]
                for iw, widget in enumerate(table_window._widgets):
                    if isinstance(widget, ptg.Label):
                        if (
                            "|" in widget.value
                            and widget.value.split("|")[1].strip().isnumeric()
                        ):
                            job_id = widget.value.split("|")[1].strip()
                            if job_id in flushable_job_ids:
                                flushable_widgets.append(widget)
                                # Jobs can be multiple rows (labels) in the table window,
                                # are underlined with a vertical line to seperate jobs
                                offset = 1
                                while (len(table_window._widgets) - (iw + offset)) > 0:
                                    current_widget = table_window._widgets[iw + offset]
                                    if not isinstance(current_widget, ptg.Label):
                                        break  # This method only cleans labels
                                    flushable_widgets.append(current_widget)
                                    if current_widget.value.startswith("+"):
                                        break  # Seperation line, stop
                                    offset += 1
                for widget in flushable_widgets:
                    table_window.remove(widget)
                manager.remove(popup)

            popup = manager.alert(
                ptg.Label("Flush non-active jobs?"),
                ptg.Button("Yes", flush),
                ptg.Button("No", lambda *_: manager.remove(popup)),
            )

        ptg.tim.define("!reload", macro_reload)
        window = ptg.Window(
            "[bold]Sparkle Jobs [!reload]%c",
            width=len(table[0]),
            box="EMPTY",
        )
        job_id_map = {job.run_id: job for job in jobs}
        for row in table:
            if "|" not in row or not row.split("|")[1].strip().isnumeric():
                window._add_widget(ptg.Label(row))
            else:
                window._add_widget(ptg.Button(label=row, onclick=cancel_jobs))

        manager.add(window)
        manager.bind(" ", flush_popup, description="Flush finished jobs")

    # If all jobs were finished, print final table.
    if all([j.status not in [Status.WAITING, Status.RUNNING] for j in jobs]):
        table = create_jobs_table(jobs, format="fancy_grid")
        print(table)


def main(argv: list[str]) -> None:
    """Main function of the jobs command."""
    # Log command call
    logging.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)

    # Filter jobs on relevant status
    path = gv.settings().DEFAULT_log_output
    jobs = [
        run
        for run in jobs_help.get_runs_from_file(path)
        if run.status == Status.WAITING or run.status == Status.RUNNING
    ]
    if args.job_ids:  # Filter
        jobs = [job for job in jobs if job.run_id in args.job_ids]
        job_ids = [job.run_id for job in jobs]
        for id in args.job_ids:
            if id not in job_ids:
                print(f"WARNING: Job ID {id} was not found ")

    if len(jobs) == 0:
        if args.job_ids:
            print(f"None of the specified jobs are running: {args.job_ids}")
            sys.exit(-1)
        print("No jobs running.")
        if args.cancel:
            sys.exit(-1)
        sys.exit(0)

    if args.cancel:
        if args.all or args.job_ids:
            killed_jobs = []
            for j in jobs:
                if args.all or j.run_id in args.job_ids:
                    j.kill()
                    killed_jobs.append(j)
            if len(killed_jobs) == 0:
                if args.all:
                    print("No jobs to cancel.")
                    sys.exit(0)
                else:
                    print(f"ERROR: No jobs with ids {args.job_ids} to cancel.")
                    sys.exit(-1)
            print(
                f"Canceled {len(killed_jobs)} jobs with IDs: "
                f"{', '.join([j.run_id for j in killed_jobs])}."
            )
        sys.exit(0)
    else:
        table_gui(jobs)
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Sparkle command to execute a parallel algorithm portfolio."""

import sys
import argparse
import random
import time
import shutil
import itertools
from operator import mod
from pathlib import Path

from tqdm import tqdm

import runrunner as rrr
from runrunner.base import Runner, Run
from runrunner.slurm import Status, SlurmRun

from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.nicknames import resolve_object_name
from sparkle.platform.settings_objects import Settings
from sparkle.solver import Solver
from sparkle.instance import Instance_Set, InstanceSet
from sparkle.types import SolverStatus, resolve_objective, UseTime
from sparkle.structures import PerformanceDataFrame


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments.

    Returns:
        parser: The parser with the parsed command line arguments
    """
    parser = argparse.ArgumentParser(
        description="Run a portfolio of solvers on an instance set in parallel."
    )
    parser.add_argument(
        *ac.InstanceSetPathsArgument.names, **ac.InstanceSetPathsArgument.kwargs
    )
    parser.add_argument(
        *ac.NicknamePortfolioArgument.names, **ac.NicknamePortfolioArgument.kwargs
    )
    parser.add_argument(*ac.SolversArgument.names, **ac.SolversArgument.kwargs)
    # Settings arguments
    parser.add_argument(*ac.SettingsFileArgument.names, **ac.SettingsFileArgument.kwargs)
    parser.add_argument(
        *Settings.OPTION_objectives.args, **Settings.OPTION_objectives.kwargs
    )
    parser.add_argument(
        *Settings.OPTION_solver_cutoff_time.args,
        **Settings.OPTION_solver_cutoff_time.kwargs,
    )
    parser.add_argument(
        *Settings.OPTION_parallel_portfolio_number_of_seeds_per_solver.args,
        **Settings.OPTION_parallel_portfolio_number_of_seeds_per_solver.kwargs,
    )
    parser.add_argument(*Settings.OPTION_run_on.args, **Settings.OPTION_run_on.kwargs)
    return parser


def create_performance_dataframe(
    solvers: list[Solver], instances_set: InstanceSet, portfolio_path: Path
) -> PerformanceDataFrame:
    """Create a PerformanceDataFrame for the given solvers and instances.

    Args:
        solvers: List of solvers to include in the PerformanceDataFrame.
        instances_set: Set of instances to include in the PerformanceDataFrame.
        portfolio_path: Path to save the CSV file.

    Returns:
        pdf: PerformanceDataFrame object initialized with solvers and instances.
    """
    instances = instances_set.instance_names
    solvers = [str(s.directory) for s in solvers]
    objectives = gv.settings().objectives
    csv_path = portfolio_path / "results.csv"
    return PerformanceDataFrame(
        csv_filepath=csv_path,
        solvers=solvers,
        objectives=objectives,
        instances=instances,
    )


def init_default_objectives() -> list:
    """Initialize default objective values and key names.

    Returns:
        default_objective_values: Dictionary with default values for each objective.
        cpu_time_key: Key for CPU time in the default values.
        status_key: Key for status in the default values.
        wall_time_key: Key for wall clock time in the default values.
    """
    # We record the 'best' of all seed results per solver-instance,
    # setting start values for objectives that are always present
    objectives = gv.settings().objectives
    cutoff = gv.settings().solver_cutoff_time
    cpu_time_key = [o.name for o in objectives if o.name.startswith("cpu_time")][0]
    status_key = [o.name for o in objectives if o.name.startswith("status")][0]
    wall_time_key = [o.name for o in objectives if o.name.startswith("wall_time")][0]
    default_objective_values = {}

    for o in objectives:
        default_value = float(sys.maxsize) if o.minimise else 0
        # Default values for time objectives can be linked to cutoff time
        if o.time and o.post_process:
            default_value = o.post_process(default_value, cutoff, SolverStatus.KILLED)
        default_objective_values[o.name] = default_value
    default_objective_values[status_key] = SolverStatus.UNKNOWN  # Overwrite status
    return default_objective_values, cpu_time_key, status_key, wall_time_key


def monitor_jobs(
    run: Run,
    instances_set: InstanceSet,
    solvers: list[Solver],
    default_objective_values: dict,
    run_on: Runner = Runner.SLURM,
) -> dict:
    """Monitor job progress and update job output dictionary.

    Args:
        run: The run object containing the submitted jobs.
        instances_set: Set of instances to run on.
        solvers: List of solvers to run on the instances.
        default_objective_values: Default objective values for each solver-instance.
        run_on: Unused

    Returns:
        job_output_dict: Dictionary containing the job output for each instance-solver
                            combination.
    """
    num_solvers, num_instances = len(solvers), len(instances_set._instance_paths)
    seeds_per_solver = gv.settings().parallel_portfolio_num_seeds_per_solver
    n_instance_jobs = num_solvers * seeds_per_solver

    job_output_dict = {
        instance_name: {
            solver.name: default_objective_values.copy() for solver in solvers
        }
        for instance_name in instances_set._instance_names
    }

    check_interval = gv.settings().parallel_portfolio_check_interval
    instances_done = [False] * num_instances

    with tqdm(total=len(instances_done)) as pbar:
        pbar.set_description("Instances done")
        while not all(instances_done):
            prev_done = sum(instances_done)
            time.sleep(check_interval)
            job_status_list = [r.status for r in run.jobs]
            job_status_completed = [
                status == Status.COMPLETED for status in job_status_list
            ]
            # The jobs are sorted by instance
            for i, instance in enumerate(instances_set._instance_paths):
                if instances_done[i]:
                    continue
                instance_job_slice = slice(
                    i * n_instance_jobs, (i + 1) * n_instance_jobs
                )
                if any(job_status_completed[instance_job_slice]):
                    instances_done[i] = True
                    # Kill remaining jobs for this instance.
                    solver_kills = [0] * num_solvers
                    for job_index in range(
                        i * n_instance_jobs, (i + 1) * n_instance_jobs
                    ):
                        if not job_status_completed[job_index]:
                            run.jobs[job_index].kill()
                            solver_index = int(
                                (mod(job_index, n_instance_jobs)) // seeds_per_solver
                            )
                            solver_kills[solver_index] += 1
                    for solver_index in range(num_solvers):
                        # All seeds of a solver were killed on instance, set status kill
                        if solver_kills[solver_index] == seeds_per_solver:
                            solver_name = solvers[solver_index].name
                            job_output_dict[instance.stem][solver_name]["status"] = (
                                SolverStatus.KILLED
                            )
            pbar.update(sum(instances_done) - prev_done)
    return job_output_dict


def wait_for_logs(cmd_list: list[str]) -> None:
    """Wait for all log files to be written.

    Args:
        cmd_list: List of command strings for all instance-solver-seed combinations.
    """
    # Attempt to verify that all logs have been written (Slurm I/O latency)
    check_interval = gv.settings().parallel_portfolio_check_interval
    for cmd in cmd_list:
        runsolver_configuration = cmd.split(" ")[:11]
        logs = [
            Path(p)
            for p in runsolver_configuration
            if Path(p).suffix in [".log", ".val", ".rawres"]
        ]
        if not all(p.exists() for p in logs):
            time.sleep(check_interval)


def update_results_from_logs(
    cmd_list: list[str],
    run: Run,
    solvers: list[Solver],
    job_output_dict: dict,
    cpu_time_key: str,
) -> dict:
    """Parse logs to update job output dictionary with best objective values.

    Args:
        cmd_list: List of command strings for all instance-solver-seed combinations.
        run: The run object containing the submitted jobs.
        solvers: List of solvers to run on the instances.
        job_output_dict: Dictionary containing the job output for each intsance-solver
                         combination.
        cpu_time_key: Key for CPU time in the job output dictionary.

    Returns:
        job_output_dict: Updated job output dictionary with best objective values.
    """
    seeds_per_solver = gv.settings().parallel_portfolio_num_seeds_per_solver
    num_solvers = len(solvers)
    n_instance_jobs = num_solvers * seeds_per_solver
    objectives = gv.settings().objectives

    for index, cmd in enumerate(cmd_list):
        solver_index = (mod(index, n_instance_jobs)) // seeds_per_solver
        solver_obj = solvers[solver_index]
        solver_output = Solver.parse_solver_output(
            run.jobs[index].stdout,
            cmd.split(" "),
            objectives=objectives,
            verifier=solver_obj.verifier,
        )
        instance_name = list(job_output_dict.keys())[index // n_instance_jobs]
        cpu_time = solver_output[cpu_time_key]
        cmd_output = job_output_dict[instance_name][solver_obj.name]
        if cpu_time > 0.0 and cpu_time < cmd_output[cpu_time_key]:
            for key, value in solver_output.items():
                if key in [o.name for o in objectives]:
                    job_output_dict[instance_name][solver_obj.name][key] = value
            if cmd_output.get("status") != SolverStatus.KILLED:
                cmd_output["status"] = solver_output.get("status")
    return job_output_dict


def fix_missing_times(
    job_output_dict: dict, status_key: str, cpu_time_key: str, wall_time_key: str
) -> dict:
    """Fix CPU and wall clock times for solvers that did not produce logs.

    Args:
        job_output_dict: Dictionary containing the job output for each instance-solver
                            combination.
        status_key: Key for status in the job output dictionary.
        cpu_time_key: Key for CPU time in the job output dictionary.
        wall_time_key: Key for wall clock time in the job output dictionary.

    Returns:
        job_output_dict: Updated job output dictionary with fixed CPU and wall clock
                            times.
    """
    cutoff = gv.settings().solver_cutoff_time
    check_interval = gv.settings().parallel_portfolio_check_interval

    # Fix the CPU/WC time for non existent logs to instance min time + check_interval
    for instance in job_output_dict.keys():
        no_log_solvers = []
        min_time = cutoff
        for solver in job_output_dict[instance].keys():
            cpu_time = job_output_dict[instance][solver][cpu_time_key]
            if cpu_time == -1.0 or cpu_time == float(sys.maxsize):
                no_log_solvers.append(solver)
            elif cpu_time < min_time:
                min_time = cpu_time
        for solver in no_log_solvers:
            job_output_dict[instance][solver][cpu_time_key] = min_time + check_interval
            job_output_dict[instance][solver][wall_time_key] = min_time + check_interval
            # Fix runtime objectives with resolved CPU/Wall times
            for key, value in job_output_dict[instance][solver].items():
                objective = resolve_objective(key)
                if objective is not None and objective.time:
                    value = (
                        job_output_dict[instance][solver][cpu_time_key]
                        if objective.use_time == UseTime.CPU_TIME
                        else job_output_dict[instance][solver][wall_time_key]
                    )
                    if objective.post_process is not None:
                        status = job_output_dict[instance][solver][status_key]
                        value = objective.post_process(value, cutoff, status)
                    job_output_dict[instance][solver][key] = value
    return job_output_dict


def print_and_write_results(
    job_output_dict: dict,
    solvers: list[Solver],
    instances_set: InstanceSet,
    portfolio_path: Path,
    status_key: str,
    cpu_time_key: str,
    wall_time_key: str,
    pdf: PerformanceDataFrame,
) -> None:
    """Print results to console and write the CSV file."""
    num_instances = len(job_output_dict)
    num_solvers = len(solvers)
    objectives = gv.settings().objectives
    for index, instance_name in enumerate(job_output_dict.keys()):
        index_str = f"[{index + 1}/{num_instances}] "
        instance_output = job_output_dict[instance_name]
        if all(
            instance_output[k][status_key] == SolverStatus.TIMEOUT
            for k in instance_output
        ):
            print(f"\n{index_str}{instance_name} was not solved within the cutoff-time.")
            continue
        print(f"\n{index_str}{instance_name} yielded the following Solver results:")
        for sindex in range(index * num_solvers, (index + 1) * num_solvers):
            solver_name = solvers[mod(sindex, num_solvers)].name
            job_info = job_output_dict[instance_name][solver_name]
            print(
                f"\t- {solver_name} ended with status {job_info[status_key]} in "
                f"{job_info[cpu_time_key]}s CPU-Time ({job_info[wall_time_key]}s "
                "Wall clock time)"
            )

    instance_map = {Path(p).name: p for p in pdf.instances}
    solver_map = {Path(s).name: s for s in pdf.solvers}
    for instance, instance_dict in job_output_dict.items():
        instance_name = Path(instance).name
        instance_full_path = instance_map.get(instance_name, instance)
        for solver, objective_dict in instance_dict.items():
            solver_name = Path(solver).name
            solver_full_path = solver_map.get(solver_name, solver)
            for objective in objectives:
                obj_name = objective.name
                obj_val = objective_dict.get(
                    obj_name, PerformanceDataFrame.missing_value
                )
                pdf.set_value(
                    value=obj_val,
                    solver=solver_full_path,
                    instance=instance_full_path,
                    objective=obj_name,
                )
    pdf.save_csv()


def build_command_list(
    instances_set: InstanceSet,
    solvers: list[Solver],
    portfolio_path: Path,
    performance_data: PerformanceDataFrame,
) -> list[str]:
    """Build the list of command strings for all instance-solver-seed combinations.

    Args:
        instances_set: Set of instances to run on.
        solvers: List of solvers to run on the instances.
        portfolio_path: Path to the parallel portfolio.
        performance_data: PerformanceDataFrame object.

    Returns:
        cmd_list: List of command strings for all instance-solver-seed combinations.
    """
    cutoff = gv.settings().solver_cutoff_time
    objectives = gv.settings().objectives
    seeds_per_solver = gv.settings().parallel_portfolio_num_seeds_per_solver
    cmd_list = []

    # Create a command for each instance-solver-seed combination
    for instance, solver in itertools.product(instances_set._instance_paths, solvers):
        for _ in range(seeds_per_solver):
            seed = int(random.getrandbits(32))
            solver_call_list = solver.build_cmd(
                instance.absolute(),
                objectives=objectives,
                seed=seed,
                cutoff_time=cutoff,
                log_dir=portfolio_path,
            )

            cmd_list.append(" ".join(solver_call_list))
            for objective in objectives:
                performance_data.set_value(
                    value=seed,
                    solver=str(solver.directory),
                    instance=instance.stem,
                    objective=objective.name,
                    solver_fields=["Seed"],
                )
    return cmd_list


def submit_jobs(
    cmd_list: list[str],
    solvers: list[Solver],
    instances_set: InstanceSet,
    run_on: Runner = Runner.SLURM,
) -> SlurmRun:
    """Submit jobs to the runner and return the run object.

    Args:
        cmd_list: List of command strings for all instance-solver-seed combinations.
        solvers: List of solvers to run on the instances.
        instances_set: Set of instances to run on.
        run_on: Runner to use for submitting the jobs.

    Returns:
        run: The run object containing the submitted jobs.
    """
    seeds_per_solver = gv.settings().parallel_portfolio_num_seeds_per_solver
    num_solvers, num_instances = len(solvers), len(instances_set._instance_paths)
    num_jobs = num_solvers * num_instances * seeds_per_solver
    parallel_jobs = min(gv.settings().slurm_jobs_in_parallel, num_jobs)
    if parallel_jobs > num_jobs:
        print(
            "WARNING: Not all jobs will be started at the same time due to the "
            "limitation of number of Slurm jobs that can be run in parallel. Check"
            " your Sparkle Slurm Settings."
        )
    print(
        f"Sparkle parallel portfolio is running {seeds_per_solver} seed(s) per solver "
        f"on {num_solvers} solvers for {num_instances} instances ..."
    )

    sbatch_options = gv.settings().sbatch_settings
    solver_names = ", ".join([s.name for s in solvers])
    # Jobs are added in to the runrunner object in the same order they are provided
    return rrr.add_to_queue(
        runner=run_on,
        cmd=cmd_list,
        name=f"Parallel Portfolio: {solver_names}",
        parallel_jobs=parallel_jobs,
        base_dir=sl.caller_log_dir,
        srun_options=["-N1", "-n1"] + sbatch_options,
        sbatch_options=sbatch_options,
        prepend=gv.settings().slurm_job_prepend,
    )


def main(argv: list[str]) -> None:
    """Main method of run parallel portfolio command."""
    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)
    settings = gv.settings(args)

    # Log command call
    sl.log_command(sys.argv, settings.random_state)
    check_for_initialise()

    # Compare current settings to latest.ini
    prev_settings = Settings(Settings.DEFAULT_previous_settings_path)
    Settings.check_settings_changes(settings, prev_settings)

    if args.solvers is not None:
        solver_paths = [
            resolve_object_name("".join(s), target_dir=settings.DEFAULT_solver_dir)
            for s in args.solvers
        ]
        if None in solver_paths:
            print("Some solvers not recognised! Check solver names:")
            for i, name in enumerate(solver_paths):
                if solver_paths[i] is None:
                    print(f'\t- "{solver_paths[i]}" ')
            sys.exit(-1)
        solvers = [Solver(p) for p in solver_paths]
    else:
        solvers = [
            Solver(p) for p in settings.DEFAULT_solver_dir.iterdir() if p.is_dir()
        ]

    portfolio_path = args.portfolio_name

    run_on = settings.run_on
    if run_on == Runner.LOCAL:
        print("Parallel Portfolio is not fully supported yet for Local runs. Exiting.")
        sys.exit(-1)

    # Retrieve instance sets
    instances = [
        resolve_object_name(
            instance_path,
            gv.file_storage_data_mapping[gv.instances_nickname_path],
            gv.settings().DEFAULT_instance_dir,
            Instance_Set,
        )
        for instance_path in args.instance_path
    ]
    # Join them into one
    if len(instances) > 1:
        print(
            "WARNING: More than one instance set specified. "
            "Currently only supporting one."
        )
    instances = instances[0]

    print(f"Running on {instances.size} instance(s)...")

    if not settings.objectives[0].time:
        print(
            "ERROR: Parallel Portfolio is currently only relevant for "
            "RunTime objectives. In all other cases, use validation"
        )
        sys.exit(-1)

    if args.portfolio_name is not None:  # Use a nickname
        portfolio_path = settings.DEFAULT_parallel_portfolio_output / args.portfolio_name
    else:  # Generate a timestamped nickname
        timestamp = time.strftime("%Y-%m-%d-%H:%M:%S", time.gmtime(time.time()))
        randintstamp = int(random.getrandbits(32))
        portfolio_path = (
            settings.DEFAULT_parallel_portfolio_output / f"{timestamp}_{randintstamp}"
        )
    if portfolio_path.exists():
        print(
            f"[WARNING] Portfolio path {portfolio_path} already exists! "
            "Overwrite? [y/n] ",
            end="",
        )
        user_input = input()
        if user_input != "y":
            sys.exit()
        shutil.rmtree(portfolio_path)

    portfolio_path.mkdir(parents=True)
    pdf = create_performance_dataframe(solvers, instances, portfolio_path)
    returned_cmd = build_command_list(instances, solvers, portfolio_path, pdf)
    default_objective_values, cpu_time_key, status_key, wall_time_key = (
        init_default_objectives()
    )
    returned_run = submit_jobs(returned_cmd, solvers, instances, Runner.SLURM)
    job_output_dict = monitor_jobs(
        returned_run, instances, solvers, default_objective_values
    )
    wait_for_logs(returned_cmd)
    job_output_dict = update_results_from_logs(
        returned_cmd, returned_run, solvers, job_output_dict, cpu_time_key
    )
    job_output_dict = fix_missing_times(
        job_output_dict, status_key, cpu_time_key, wall_time_key
    )
    print_and_write_results(
        job_output_dict,
        solvers,
        instances,
        portfolio_path,
        status_key,
        cpu_time_key,
        wall_time_key,
        pdf,
    )

    # Write used settings to file
    settings.write_used_settings()
    print("Running Sparkle parallel portfolio is done!")
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
"""Definitions of constants broadly used in Sparkle."""

import ast
from argparse import Namespace
import random
import numpy as np

from sparkle.platform.settings_objects import Settings
from sparkle.structures import PerformanceDataFrame
from sparkle.configurator.configurator import ConfigurationScenario
from sparkle.configurator.implementations import (
    SMAC2Scenario,
    SMAC3Scenario,
    ParamILSScenario,
    IRACEScenario,
)
from sparkle.selector import SelectionScenario


__settings: Settings = None


def settings(argsv: Namespace = None) -> Settings:
    """Function to get the global settings object."""
    global __settings
    if __settings is None:
        __settings = Settings(Settings.DEFAULT_settings_path, argsv=argsv)
        # Set global random state
        max_seed = 2**32 - 1
        latest_ini = Settings(Settings.DEFAULT_previous_settings_path)
        # Determine seed priority: latest.ini > __settings > random
        seed = latest_ini.seed or __settings.seed or random.randint(0, max_seed)
        # Set global RNG states
        np.random.seed(seed)
        random.seed(seed)
        __settings.random_state = seed
        # Next seed will be saved in latest.ini when the cli script calls 'write_used_settings()'
        next_seed = random.randint(0, max_seed)
        __settings.seed = next_seed

    elif argsv is not None:
        __settings.apply_arguments(argsv)

    return __settings


__configuration_scenarios: list[ConfigurationScenario] = None
__selection_scenarios: list[SelectionScenario] = None


def configuration_scenarios(refresh: bool = False) -> list[ConfigurationScenario]:
    """Fetch all known configuration scenarios."""
    global __configuration_scenarios
    config_path = Settings.DEFAULT_configuration_output
    if __configuration_scenarios is None or refresh:
        __configuration_scenarios = []
        for f in config_path.glob("*/*/*.*"):  # We look for files at depth three
            if "scenario" not in f.name:
                continue
            if "SMAC2" in str(f):
                __configuration_scenarios.append(SMAC2Scenario.from_file(f))
            elif "SMAC3" in str(f):
                __configuration_scenarios.append(SMAC3Scenario.from_file(f))
            elif "ParamILS" in str(f):
                __configuration_scenarios.append(ParamILSScenario.from_file(f))
            elif "IRACE" in str(f):
                __configuration_scenarios.append(IRACEScenario.from_file(f))
    return __configuration_scenarios


def selection_scenarios(refresh: bool = False) -> list[SelectionScenario]:
    """Fetch all known selection scenarios."""
    global __selection_scenarios
    selection_path = Settings.DEFAULT_selection_output
    if __selection_scenarios is None or refresh:
        __selection_scenarios = []
        for f in selection_path.glob("*/*/*.txt"):  # We look for files at depth three
            if "scenario" not in f.name:
                continue
            __selection_scenarios.append(SelectionScenario.from_file(f))
    return __selection_scenarios


def parallel_portfolio_scenarios() -> list[PerformanceDataFrame]:
    """Fetch all known parallel portfolio scenarios."""
    parallel_portfolio_path = Settings.DEFAULT_parallel_portfolio_output
    return [PerformanceDataFrame(f) for f in parallel_portfolio_path.glob("*/*.csv")]


reference_list_dir = Settings.DEFAULT_reference_dir
extractor_nickname_list_path = reference_list_dir / "sparkle_extractor_nickname_list.txt"
solver_nickname_list_path = reference_list_dir / "sparkle_solver_nickname_list.txt"
instances_nickname_path = reference_list_dir / "sparkle_instance_nickname_list.txt"

file_storage_data_mapping = {
    solver_nickname_list_path: {},
    instances_nickname_path: {},
    extractor_nickname_list_path: {},
}

for data_path in file_storage_data_mapping.keys():
    if data_path.exists():
        with data_path.open("r+") as fo:
            file_storage_data_mapping[data_path] = ast.literal_eval(fo.read())

solver_nickname_mapping = file_storage_data_mapping[solver_nickname_list_path]
instance_set_nickname_mapping = file_storage_data_mapping[instances_nickname_path]



"""This package provides instance set support for Sparkle."""

from sparkle.instance.instances import (
    MultiFileInstanceSet,
    FileInstanceSet,
    IterableFileInstanceSet,
    InstanceSet,
)
from pathlib import Path


def Instance_Set(target: any) -> InstanceSet:
    """The combined interface for all instance set types."""
    if (
        isinstance(target, Path)
        and (target / MultiFileInstanceSet.instance_csv).exists()
    ) or (
        isinstance(target, list)
        and isinstance(target[0], Path)
        and (target[0].parent / MultiFileInstanceSet.instance_csv).exists()
    ):
        return MultiFileInstanceSet(target)
    elif (not target.exists()) and (
        target.parent / MultiFileInstanceSet.instance_csv
    ).exists():
        # Single instance
        return MultiFileInstanceSet(target)
    elif (
        isinstance(target, Path)
        and target.is_dir()
        and all(
            [
                p.suffix in IterableFileInstanceSet.supported_filetypes
                for p in target.iterdir()
            ]
        )
    ):
        return IterableFileInstanceSet(target)
    elif not target.exists():  # Resolve suffix
        alternatives = [p for p in target.parent.iterdir()]
        for alt in alternatives:
            if target.name == alt.stem:
                target = alt
                break
    return FileInstanceSet(target)
